from gi.repository import Gtk

from .terminal import generate_xrdb_theme_from_oomox
from .color import convert_theme_color_to_gdk
from .i18n import _


WIDGET_SPACING = 10


class TerminalThemePreview(Gtk.Grid):

    LEFT_MARGIN = 18

    COLOR_ROWS = (
        (_("black"), 0, 8),
        (_("red"), 1, 9),
        (_("green"), 2, 10),
        (_("yellow"), 3, 11),
        (_("blue"), 4, 12),
        (_("purple"), 5, 13),
        (_("cyan"), 6, 14),
        (_("white"), 7, 15),
    )
    terminal_widgets = None

    def __init__(self):
        super().__init__(row_spacing=6, column_spacing=6)
        self.set_margin_left(WIDGET_SPACING)
        self.set_margin_right(WIDGET_SPACING)

        self.background = Gtk.Grid(row_spacing=6, column_spacing=6)
        self.background.set_margin_top(WIDGET_SPACING/2)
        self.background.set_margin_bottom(WIDGET_SPACING)
        self.background.set_margin_right(WIDGET_SPACING)

        self.terminal_widgets = {}
        twi = self.terminal_widgets

        twi["normal"] = Gtk.Label()
        twi["normal"].set_markup("<tt>{}</tt>".format(_("terminal colors:")))
        self.background.attach(twi["normal"], 1, 1, 2, 1)
        previous_row = twi["normal"]
        previous_row.set_margin_left(self.LEFT_MARGIN)
        for color_row in self.COLOR_ROWS:
            color_name, normal_id, highlight_id = color_row
            key1 = "color{}".format(normal_id)
            key2 = "color{}".format(highlight_id)
            twi[key1] = Gtk.Label()
            twi[key2] = Gtk.Label()
            twi[key1].set_markup("<tt>{}</tt>".format(color_name))
            twi[key2].set_markup("<tt>{}</tt>".format(color_name))
            self.background.attach_next_to(
                twi[key1], previous_row,
                Gtk.PositionType.BOTTOM, 1, 1
            )
            self.background.attach_next_to(
                twi[key2], twi[key1],
                Gtk.PositionType.RIGHT, 1, 1
            )
            previous_row = twi[key1]
            previous_row.set_margin_left(self.LEFT_MARGIN)
        self.attach(self.background, 1, 1, 1, 1)

    def update_preview(self, colorscheme):
        term_colorscheme = generate_xrdb_theme_from_oomox(colorscheme)
        converted = {
            key: convert_theme_color_to_gdk(theme_value)
            for key, theme_value in term_colorscheme.items()
        }
        term_bg = converted["background"]
        self.terminal_widgets["normal"].override_color(
            Gtk.StateType.NORMAL, converted["foreground"]
        )
        self.override_background_color(Gtk.StateType.NORMAL, term_bg)
        for color_row in self.COLOR_ROWS:
            _color_name, normal_id, highlight_id = color_row
            key1 = "color{}".format(normal_id)
            key2 = "color{}".format(highlight_id)
            self.terminal_widgets[key1].override_color(
                Gtk.StateType.NORMAL, converted[key1]
            )
            self.terminal_widgets[key2].override_color(
                Gtk.StateType.NORMAL, term_bg
            )
            self.terminal_widgets[key2].override_background_color(
                Gtk.StateType.NORMAL, converted[key2]
            )



import random

from gi.repository import Gdk

from .config import FALLBACK_COLOR


def hex_to_int(text):
    return int("0x{}".format(text), 0)


def int_to_hex(myint):
    return "{0:02x}".format(int(myint))


def color_list_from_hex(color_text):
    return [color_text[:2], color_text[2:4], color_text[4:]]


def color_hex_from_list(color_list):
    return ''.join([int_to_hex(i) for i in color_list])


def is_dark(color_text):
    return sum([
        hex_to_int(channel_text)
        for channel_text in color_list_from_hex(color_text)
    ]) < 384


def hex_darker(color_text, darken_amount=10):
    return color_hex_from_list([
        max(min(hex_to_int(channel_text) - darken_amount, 255), 0)
        for channel_text in color_list_from_hex(color_text)
    ])


class ColorDiff():
    r = None  # pylint: disable=invalid-name
    g = None  # pylint: disable=invalid-name
    b = None  # pylint: disable=invalid-name

    @property
    def list(self):
        return [self.r, self.g, self.b]

    @property
    def abs_list(self):
        return [abs(c) for c in self.list]

    @property
    def min(self):
        return min(self.abs_list)

    @property
    def abs(self):
        return sum(self.abs_list)

    @property
    def minabs(self):
        return self.min + self.abs

    @property
    def avg(self):
        return self.abs / 3

    @property
    def sat(self):
        r, g, b = self.abs_list  # pylint: disable=invalid-name
        return abs(r-g)+abs(r-b) + abs(g-r)+abs(g-b) + abs(b-g)+abs(b-r)

    def __repr__(self):
        return str(self.abs)

    channels = ['r', 'g', 'b']

    def __init__(self, theme_color_1, theme_color_2):
        color_list_1 = color_list_from_hex(theme_color_1)
        color_list_2 = color_list_from_hex(theme_color_2)
        for channel_index, channel_1_text in enumerate(color_list_1):
            channel_1 = hex_to_int(channel_1_text)
            channel_2 = hex_to_int(color_list_2[channel_index])
            setattr(self, self.channels[channel_index], channel_1-channel_2)

    def apply_to(self, color_text):
        color_list = color_list_from_hex(color_text)
        result = ''
        for channel_index, channel_text in enumerate(color_list):
            channel = hex_to_int(channel_text)
            int_result = channel - getattr(self, self.channels[channel_index])
            if int_result < 0:
                int_result = 0
            if int_result > 255:
                int_result = 255
            result += int_to_hex(int_result)
        return result


SMALLEST_DIFF = ColorDiff("000000", "ffffff")


def find_closest_color(color_hex, colors_hex):
    smallest_diff = SMALLEST_DIFF
    closest_color = None
    for preset_color in colors_hex:
        diff = ColorDiff(preset_color, color_hex)
        if diff.abs < smallest_diff.abs:
            smallest_diff = diff
            closest_color = preset_color
    return closest_color, smallest_diff


def convert_theme_color_to_gdk(theme_color):
    gdk_color = Gdk.RGBA()
    gdk_color.parse("#" + theme_color)
    return gdk_color


def convert_gdk_to_theme_color(gdk_color):
    return "".join([
        int_to_hex(n * 255)
        for n in (gdk_color.red, gdk_color.green, gdk_color.blue)
    ])


def get_random_gdk_color():
    return Gdk.RGBA(random.random(), random.random(), random.random(), 1)


def get_random_theme_color():
    return convert_gdk_to_theme_color(get_random_gdk_color())


def mix_gdk_colors(gdk_color_1, gdk_color_2, ratio):
    result_gdk_color = Gdk.RGBA()
    for attr in ('red', 'green', 'blue', 'alpha'):
        setattr(
            result_gdk_color, attr,
            (getattr(gdk_color_1, attr) * ratio + getattr(gdk_color_2, attr) *
             (1 - ratio)))
    return result_gdk_color


def mix_theme_colors(theme_color_1, theme_color_2, ratio):
    color_list_1 = []
    color_list_2 = []
    for color_text, color_list in ((theme_color_1, color_list_1),
                                   (theme_color_2, color_list_2)):
        color_list.append(color_text[:2])
        color_list.append(color_text[2:4])
        color_list.append(color_text[4:])
    result = ''
    for channel_index, channel_1_text in enumerate(color_list_1):
        try:
            channel_1 = hex_to_int(channel_1_text)
            channel_2 = hex_to_int(color_list_2[channel_index])
        except ValueError:
            return FALLBACK_COLOR
        result += int_to_hex(channel_1 * ratio + channel_2 * (1 - ratio))
    return result



from abc import ABCMeta, abstractproperty, abstractmethod
from enum import Enum

from oomox_gui.config import FALLBACK_COLOR


class OomoxPlugin(metaclass=ABCMeta):

    @abstractproperty
    def name(self):
        pass

    @abstractproperty
    def display_name(self):
        pass


class OomoxThemePlugin(OomoxPlugin):

    @abstractproperty
    def description(self):
        pass

    @abstractproperty
    def export_dialog(self):
        pass

    @abstractproperty
    def gtk_preview_dir(self):
        pass

    enabled_keys_gtk = []
    enabled_keys_options = []
    enabled_keys_other = []
    enabled_keys_extra = []

    theme_model_gtk = []
    theme_model_options = []
    theme_model_other = []
    theme_model_extra = []

    def preview_before_load_callback(self, preview_object, colorscheme):
        pass

    class PreviewImageboxesNames(Enum):
        CHECKBOX = 'checkbox-checked'

    preview_sizes = {
        PreviewImageboxesNames.CHECKBOX.name: 16,
    }

    def preview_transform_function(self, svg_template, colorscheme):
        # pylint: disable=no-self-use
        for key in (
                "SEL_BG", "ACCENT_BG", "TXT_BG", "BG", "FG",
        ):
            svg_template = svg_template.replace(
                "%{}%".format(key), colorscheme.get(key) or FALLBACK_COLOR
            )
        return svg_template


class OomoxIconsPlugin(OomoxPlugin):

    enabled_keys_icons = []
    theme_model_icons = []

    @abstractproperty
    def preview_svg_dir(self):
        pass

    @abstractmethod
    def preview_transform_function(self, svg_template, colorscheme):
        pass


class OomoxExportPlugin(OomoxPlugin):

    @abstractproperty
    def export_dialog(self):
        pass

    theme_model_extra = []


class OomoxImportPlugin(OomoxPlugin):

    # @abstractproperty
    # def import_dialog(self):
        # pass

    @abstractproperty
    def file_extensions(self):
        pass

    @abstractmethod
    def read_colorscheme_from_path(self, preset_path):
        pass

    theme_model_import = []



import os
import gettext


LOCALEDIR = os.path.join(
    os.path.dirname(os.path.realpath(__file__)),
    '..', 'locale'
)
TRANSLATION = gettext.translation(
    'oomox',
    localedir=LOCALEDIR,
    fallback=True
)


def _(msg: str) -> str:
    return TRANSLATION.gettext(msg)


def _n(singular: str, plural: str, count: int) -> str:
    return TRANSLATION.ngettext(singular, plural, count)



import os

from gi.repository import Gtk, GLib

from .theme_model import THEME_MODEL
from .color import (
    convert_theme_color_to_gdk, mix_theme_colors, mix_gdk_colors,
)
from .gtk_helpers import ScaledImage
from .preview_terminal import TerminalThemePreview
from .preview_icons import IconThemePreview
from .config import FALLBACK_COLOR
from .i18n import _


WIDGET_SPACING = 10


class CssProviders():
    theme = None
    gradient = None
    border = None
    headerbar_border = None
    wm_border = None
    caret = None
    reset_style = None

    def __init__(self):
        self.theme = {}
        self.gradient = {}
        self.border = {}
        self.headerbar_border = Gtk.CssProvider()
        self.headerbar_border.load_from_data((
            """
            headerbar {
                border: none;
            }
            """
        ).encode('ascii'))
        self.wm_border = Gtk.CssProvider()
        self.caret = Gtk.CssProvider()
        self.reset_style = Gtk.CssProvider()
        self.reset_style.load_from_data((
            """
            * {
                box-shadow:none;
                border: none;
            }
            """
        ).encode('ascii'))


class PreviewHeaderbar(Gtk.HeaderBar):

    title = None
    button = None

    def __init__(self):
        super().__init__()
        self.set_show_close_button(False)
        self.title = Gtk.Label(label=_("Headerbar"))
        self.props.custom_title = self.title
        self.button = Gtk.Button(label="   %s   " % _("Button"))
        self.pack_end(self.button)


class PreviewWidgets():

    # gtk preview widgets:
    background = None
    headerbar = None
    menubar = None
    label = None
    sel_label = None
    entry = None
    preview_imageboxes = None
    preview_imageboxes_templates = None
    button = None

    def __init__(self):
        self.background = Gtk.Grid(row_spacing=6, column_spacing=6)

        headerbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)

        self.headerbar = PreviewHeaderbar()

        self.menubar = Gtk.MenuBar()
        menuitem1 = Gtk.MenuItem(label=_('File'))
        menuitem1.set_submenu(self.create_menu(3, True))
        self.menubar.append(menuitem1)
        menuitem2 = Gtk.MenuItem(label=_('Edit'))
        menuitem2.set_submenu(self.create_menu(6, True))
        self.menubar.append(menuitem2)

        headerbox.pack_start(self.headerbar, True, True, 0)
        headerbox.pack_start(self.menubar, True, True, 0)

        self.label = Gtk.Label(label=_("This is a label."))
        self.sel_label = Gtk.Label(label=_("Selected item."))
        self.entry = Gtk.Entry(text=_("Text entry."))
        self.button = Gtk.Button(label=_("Click-click"))

        self.preview_imageboxes = {}
        self.preview_imageboxes_templates = {}
        self.preview_imageboxes['CHECKBOX'] = ScaledImage(width=16)

        fake_checkbox = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=6)
        fake_checkbox.pack_start(self.preview_imageboxes['CHECKBOX'], False, False, 0)
        fake_checkbox.pack_start(self.label, False, False, 0)
        fake_checkbox.set_margin_left(20)

        self.background.attach(headerbox, 1, 1, 5, 2)
        self.background.attach(fake_checkbox, 3, 3, 1, 1)
        self.background.attach_next_to(
            self.sel_label, fake_checkbox, Gtk.PositionType.BOTTOM, 1, 1
        )
        self.background.attach_next_to(
            self.entry, self.sel_label, Gtk.PositionType.BOTTOM, 1, 1
        )
        self.background.attach_next_to(
            self.button, self.entry, Gtk.PositionType.BOTTOM, 1, 1
        )

    def create_menu(self, n_items, has_submenus=False):
        menu = Gtk.Menu()
        for i in range(0, n_items):
            sensitive = (i + 1) % 3 != 0
            label = _('Item {id}') if sensitive else _('Insensitive Item {id}')
            item = Gtk.MenuItem(label=label.format(id=i + 1),
                                sensitive=sensitive)
            menu.append(item)
            if has_submenus and (i + 1) % 2 == 0:
                item.set_submenu(self.create_menu(2))
        return menu

    def load_imageboxes_templates(self, theme_plugin):
        for icon in theme_plugin.PreviewImageboxesNames:
            template_path = "{}.svg.template".format(icon.value)
            with open(
                    os.path.join(
                        theme_plugin.gtk_preview_dir, template_path
                    ), "rb"
            ) as file_object:
                self.preview_imageboxes_templates[icon.name] = file_object.read().decode('utf-8')

    def update_preview_imageboxes(self, colorscheme, theme_plugin):
        transform_function = theme_plugin.preview_transform_function
        self.load_imageboxes_templates(theme_plugin)
        for icon in theme_plugin.PreviewImageboxesNames:
            new_svg_image = transform_function(
                self.preview_imageboxes_templates[icon.name],
                colorscheme
            ).encode('ascii')
            self.preview_imageboxes[icon.name].set_from_bytes(
                new_svg_image, width=theme_plugin.preview_sizes[icon.name]
            )


def _get_menu_widgets(shell):
    """ gets a menu shell (menu or menubar) and all its children """
    children = [shell]
    for child in shell:
        children.append(child)
        submenu = child.get_submenu()
        if submenu:
            children.extend(_get_menu_widgets(submenu))
    return children


class ThemePreview(Gtk.Grid):

    BG = 'bg'  # pylint: disable=invalid-name
    FG = 'fg'  # pylint: disable=invalid-name

    WM_BORDER_WIDTH = 2

    theme_plugin_name = None
    widgets = None
    css_providers = None

    # widget sections:
    background = None
    gtk_preview = None
    icons_preview = None
    terminal_preview = None

    def __init__(self):
        super().__init__(row_spacing=6, column_spacing=6)
        self.css_providers = CssProviders()
        self.init_widgets()

    def init_widgets(self):
        self.gtk_preview = PreviewWidgets()
        self.background = Gtk.Grid(row_spacing=WIDGET_SPACING, column_spacing=6)
        preview_label = Gtk.Label(label=_("Preview:"))
        self.attach(preview_label, 1, 1, 3, 1)
        self.attach_next_to(self.background, preview_label,
                            Gtk.PositionType.BOTTOM, 1, 1)

        self.gtk_preview.background.set_margin_bottom(WIDGET_SPACING)
        self.background.attach(self.gtk_preview.background, 1, 3, 1, 1)

        if self.icons_preview:
            self.icons_preview.destroy()
        self.icons_preview = IconThemePreview()
        self.background.attach_next_to(
            self.icons_preview, self.gtk_preview.background,
            Gtk.PositionType.BOTTOM, 1, 1
        )

        if self.terminal_preview:
            self.terminal_preview.destroy()
        self.terminal_preview = TerminalThemePreview()
        self.terminal_preview.set_margin_bottom(WIDGET_SPACING)
        self.background.attach_next_to(
            self.terminal_preview, self.icons_preview,
            Gtk.PositionType.BOTTOM, 1, 1
        )
        self.background.set_margin_bottom(WIDGET_SPACING)

        self.gtk_preview.button.connect("style-updated", self._queue_resize)

    def override_widget_color(self, widget, value, color, state=Gtk.StateType.NORMAL):
        if value == self.BG:
            return widget.override_background_color(state, color)
        if value == self.FG:
            return widget.override_color(state, color)
        raise NotImplementedError()

    def update_preview_carets(self, colorscheme):
        self.css_providers.caret.load_from_data((
            (Gtk.get_minor_version() >= 20 and """
            * {{
                caret-color: #{primary_caret_color};
                -gtk-secondary-caret-color: #{secondary_caret_color};
                -GtkWidget-cursor-aspect-ratio: {caret_aspect_ratio};
            }}
            """ or """
            * {{
                -GtkWidget-cursor-color: #{primary_caret_color};
                -GtkWidget-secondary-cursor-color: #{secondary_caret_color};
                -GtkWidget-cursor-aspect-ratio: {caret_aspect_ratio};
            }}
            """).format(
                primary_caret_color=colorscheme['CARET1_FG'],
                secondary_caret_color=colorscheme['CARET2_FG'],
                caret_aspect_ratio=colorscheme['CARET_SIZE']
            )
        ).encode('ascii'))
        Gtk.StyleContext.add_provider(
            self.gtk_preview.entry.get_style_context(),
            self.css_providers.caret,
            Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION
        )

    def update_preview_gradients(self, colorscheme):
        gradient = colorscheme['GRADIENT']
        if gradient == 0:
            self.reset_gradients()
            return
        for widget, color_key in zip(
                [
                    self.gtk_preview.button,
                    self.gtk_preview.headerbar.button,
                    self.gtk_preview.entry,
                    self.gtk_preview.headerbar,
                ],
                [
                    "BTN_BG",
                    "HDR_BTN_BG",
                    "TXT_BG",
                    "MENU_BG"
                ]
        ):
            color = colorscheme[color_key]
            css_provider_gradient = self.css_providers.gradient.get(color_key)
            if not css_provider_gradient:
                css_provider_gradient = \
                        self.css_providers.gradient[color_key] = \
                        Gtk.CssProvider()
            css_provider_gradient.load_from_data((
                """
                * {{
                    background-image: linear-gradient(to bottom,
                        shade(#{color}, {amount1}),
                        shade(#{color}, {amount2})
                    );
                }}
                """.format(
                    color=color,
                    amount1=1 - gradient / 2,
                    amount2=1 + gradient * 2
                )
            ).encode('ascii'))
            Gtk.StyleContext.add_provider(
                widget.get_style_context(),
                css_provider_gradient,
                Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION
            )

    def reset_gradients(self):
        css_provider_gradient = self.css_providers.gradient.get("reset")
        if not css_provider_gradient:
            css_provider_gradient = \
                    self.css_providers.gradient["reset"] = \
                    Gtk.CssProvider()
            css_provider_gradient.load_from_data((
                """
                * {
                    background-image: none;
                }
                """
            ).encode('ascii'))
        for widget in [
                self.gtk_preview.button,
                self.gtk_preview.headerbar.button,
                self.gtk_preview.entry,
                self.gtk_preview.headerbar,
        ]:
            Gtk.StyleContext.add_provider(
                widget.get_style_context(),
                css_provider_gradient,
                Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION
            )

    def update_preview_borders(self, colorscheme):
        for widget_name, widget, fg, bg, ratio in (  # pylint: disable=invalid-name
                (
                    'button',
                    self.gtk_preview.button,
                    colorscheme['BTN_FG'],
                    colorscheme['BTN_BG'],
                    0.22
                ), (
                    'headerbar_button',
                    self.gtk_preview.headerbar.button,
                    colorscheme['HDR_BTN_FG'],
                    colorscheme['HDR_BTN_BG'],
                    0.22
                ), (
                    'entry',
                    self.gtk_preview.entry,
                    mix_theme_colors(
                        colorscheme['TXT_FG'], colorscheme['TXT_BG'], 0.20
                    ),
                    colorscheme['BG'],
                    0.69
                ),
        ):
            border_color = mix_theme_colors(fg, bg, ratio)
            css_provider_border_color = self.css_providers.border.get(widget_name)
            if not css_provider_border_color:
                css_provider_border_color = \
                    self.css_providers.border[widget_name] = \
                    Gtk.CssProvider()
            css_provider_border_color.load_from_data(
                """
                * {{
                    border-color: #{border_color};
                    border-radius: {roundness}px;
                }}
                """.format(
                    border_color=border_color,
                    roundness=colorscheme["ROUNDNESS"],
                ).encode('ascii')
            )
            Gtk.StyleContext.add_provider(
                widget.get_style_context(),
                css_provider_border_color,
                Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION
            )

    def update_preview_colors(self, colorscheme):

        converted = {
            theme_value['key']: convert_theme_color_to_gdk(
                colorscheme[theme_value['key']]
            )
            for theme_value in THEME_MODEL if (
                theme_value['type'] == 'color' and
                not theme_value['key'].startswith('TERMINAL_')
            )
        }

        def mix(color1, color2, amount):
            return mix_gdk_colors(converted[color2], converted[color1], amount)

        self.override_widget_color(self.background, self.BG, converted["BG"])
        self.override_widget_color(self.gtk_preview.label, self.FG, converted["FG"])
        self.override_widget_color(self.gtk_preview.sel_label, self.FG, converted["SEL_FG"])
        self.override_widget_color(self.gtk_preview.sel_label, self.BG, converted["SEL_BG"])
        self.override_widget_color(self.gtk_preview.entry, self.FG, converted["TXT_FG"])
        self.override_widget_color(self.gtk_preview.entry, self.BG, converted["TXT_BG"])
        self.override_widget_color(
            self.gtk_preview.entry, self.FG, converted["SEL_FG"],
            state=Gtk.StateFlags.SELECTED
        )
        self.override_widget_color(
            self.gtk_preview.entry, self.BG, converted["SEL_BG"],
            state=Gtk.StateFlags.SELECTED
        )
        self.override_widget_color(self.gtk_preview.button, self.FG, converted["BTN_FG"])
        self.override_widget_color(self.gtk_preview.button, self.BG, converted["BTN_BG"])
        for item in self.gtk_preview.menubar.get_children():
            self.override_widget_color(item, self.FG, converted["MENU_FG"])
            self.override_widget_color(
                item, self.BG, mix("MENU_BG", "MENU_FG", 0.21),
                state=Gtk.StateFlags.PRELIGHT
            )
            for widget in _get_menu_widgets(item.get_submenu()):
                if isinstance(widget, Gtk.MenuShell):
                    self.override_widget_color(widget, self.BG, converted["MENU_BG"])
                    self.override_widget_color(widget, self.FG, converted["MENU_FG"])
                else:
                    if not widget.get_sensitive():  # :disabled
                        color = mix("MENU_FG", "MENU_BG", 0.5)
                    else:
                        color = converted["MENU_FG"]
                    self.override_widget_color(widget, self.FG, color)
                self.override_widget_color(
                    widget, self.BG, converted["SEL_BG"],
                    state=Gtk.StateFlags.PRELIGHT
                )
                self.override_widget_color(
                    widget, self.FG, converted["SEL_FG"],
                    state=Gtk.StateFlags.PRELIGHT
                )
        self.override_widget_color(self.gtk_preview.menubar, self.BG, converted["MENU_BG"])
        self.override_widget_color(self.gtk_preview.headerbar, self.BG, converted["MENU_BG"])
        self.override_widget_color(self.gtk_preview.headerbar.title, self.FG, converted["MENU_FG"])
        self.override_widget_color(
            self.gtk_preview.headerbar.button, self.FG,
            converted["HDR_BTN_FG"]
        )
        self.override_widget_color(
            self.gtk_preview.headerbar.button, self.BG,
            converted["HDR_BTN_BG"]
        )

        self.override_widget_color(
            self.icons_preview, self.BG,
            converted["TXT_BG"]
        )

        self.css_providers.wm_border.load_from_data("""
            * {{
                border-color: #{border_color};
                /*border-radius: {roundness}px;*/
                border-width: {wm_border_width}px;
                border-style: solid;
            }}
        """.format(
            border_color=colorscheme['WM_BORDER_FOCUS'],
            roundness=colorscheme['ROUNDNESS'],
            wm_border_width=self.WM_BORDER_WIDTH
        ).encode('ascii'))
        Gtk.StyleContext.add_provider(
            self.background.get_style_context(),
            self.css_providers.wm_border,
            Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION
        )

    def update_preview(self, colorscheme, theme_plugin, icons_plugin):
        if not theme_plugin:
            return
        _colorscheme = colorscheme
        colorscheme = {}
        colorscheme.update(_colorscheme)
        theme_plugin.preview_before_load_callback(self, colorscheme)

        colorscheme_with_fallbacks = {}
        for theme_value in THEME_MODEL:
            if 'key' not in theme_value:
                continue
            result = colorscheme.get(theme_value['key'])
            if not result and theme_value['type'] == 'color':
                result = FALLBACK_COLOR
            colorscheme_with_fallbacks[theme_value['key']] = result

        self.override_css_style(colorscheme_with_fallbacks, theme_plugin)
        self.update_preview_colors(colorscheme_with_fallbacks)
        self.update_preview_borders(colorscheme_with_fallbacks)
        self.update_preview_carets(colorscheme_with_fallbacks)
        self.update_preview_gradients(colorscheme_with_fallbacks)
        self.gtk_preview.update_preview_imageboxes(colorscheme_with_fallbacks, theme_plugin)

        self.icons_preview.update_preview(colorscheme, icons_plugin)
        self.terminal_preview.update_preview(colorscheme)

    def get_theme_css_provider(self, theme_plugin):
        css_name = "theme{}.css".format(
            '20' if Gtk.get_minor_version() >= 20 else ''
        )
        css_dir = theme_plugin.gtk_preview_dir
        css_path = os.path.join(css_dir, css_name)
        css_provider = self.css_providers.theme.get(css_path)
        if css_provider:
            return css_provider
        css_provider = Gtk.CssProvider()
        try:
            css_provider.load_from_path(css_path)
        except GLib.Error as exc:  # pylint: disable=catching-non-exception
            print(exc)
        self.css_providers.theme[css_path] = css_provider
        return css_provider

    def override_css_style(self, colorscheme, theme_plugin):
        new_theme_plugin_name = colorscheme["THEME_STYLE"]
        if new_theme_plugin_name == self.theme_plugin_name:
            return
        if self.theme_plugin_name:
            for child in self.get_children():
                self.remove(child)
                child.destroy()
            self.init_widgets()
        self.theme_plugin_name = new_theme_plugin_name
        base_theme_css_provider = self.get_theme_css_provider(theme_plugin)

        def apply_css(widget):
            widget_style_context = widget.get_style_context()
            Gtk.StyleContext.add_provider(
                widget_style_context,
                self.css_providers.reset_style,
                Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION
            )
            Gtk.StyleContext.add_provider(
                widget_style_context,
                base_theme_css_provider,
                Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION
            )
            if isinstance(widget, Gtk.Container):
                widget.forall(apply_css)
        apply_css(self)

        Gtk.StyleContext.add_provider(
            self.gtk_preview.headerbar.get_style_context(),
            self.css_providers.headerbar_border,
            Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION
        )

        self.show_all()

    def _queue_resize(self, *_args):
        # print(args)
        self.queue_resize()



import os

from .config import TERMINAL_TEMPLATE_DIR
from .plugin_loader import (
    THEME_PLUGINS, ICONS_PLUGINS, EXPORT_PLUGINS, IMPORT_PLUGINS,
)
from .i18n import _


def get_base_keys(base_theme_model):
    return {
        theme_value['key']: index
        for index, theme_value in enumerate(base_theme_model)
        if 'key' in theme_value
    }


def merge_model_with_base(
        whole_theme_model, plugin_model_name,
        plugins, base_theme_model=None, value_filter_key=None
):
    if base_theme_model is None:
        base_theme_model = []
    base_keys = get_base_keys(base_theme_model)

    if value_filter_key:
        for theme_value in base_theme_model:
            if 'key' in theme_value:
                theme_value.setdefault('value_filter', {}).setdefault(value_filter_key, [])

    for theme_plugin in plugins.values():
        plugin_theme_model = getattr(theme_plugin, "theme_model_"+plugin_model_name)
        for theme_value in plugin_theme_model:
            if 'key' not in theme_value or theme_value['key'] not in base_keys:
                base_theme_model.append(theme_value)
                base_keys = get_base_keys(base_theme_model)
        plugin_theme_model_keys = [
            theme_value['key']
            for theme_value in plugin_theme_model
            if 'key' in theme_value
        ]
        plugin_enabled_keys = getattr(
            theme_plugin, "enabled_keys_"+plugin_model_name, []
        )
        if not value_filter_key:
            continue
        for key in plugin_theme_model_keys + plugin_enabled_keys:
            base_theme_value = base_theme_model[base_keys[key]]
            value_filter = base_theme_value.setdefault('value_filter', {})
            value_filter_theme_style = value_filter.setdefault(value_filter_key, [])
            if not isinstance(value_filter_theme_style, list):
                value_filter_theme_style = [value_filter_theme_style, ]
            value_filter_theme_style.append(theme_plugin.name)
            base_theme_value['value_filter'][value_filter_key] = value_filter_theme_style
    whole_theme_model += base_theme_model


def merge_theme_model_with_base(whole_theme_model, base_theme_model, plugin_model_name):
    return merge_model_with_base(
        whole_theme_model=whole_theme_model,
        base_theme_model=base_theme_model,
        plugin_model_name=plugin_model_name,
        value_filter_key='THEME_STYLE',
        plugins=THEME_PLUGINS,
    )


THEME_MODEL = []
merge_model_with_base(
    whole_theme_model=THEME_MODEL,
    plugin_model_name='import',
    value_filter_key='FROM_PLUGIN',
    plugins=IMPORT_PLUGINS,
)
THEME_MODEL = THEME_MODEL + [
    {
        'key': 'THEME_STYLE',
        'type': 'options',
        'options': [
            {
                'value': theme_plugin.name,
                'display_name': theme_plugin.display_name,
                'description': theme_plugin.description,
            }
            for theme_plugin in THEME_PLUGINS.values()
        ],
        'fallback_value': 'oomox',
        'display_name': _('Theme style'),
    },
]

BASE_THEME_MODEL_GTK = [
    {
        'key': 'BG',
        'type': 'color',
        'display_name': _('Background')
    },
    {
        'key': 'FG',
        'type': 'color',
        'display_name': _('Foreground/text')
    },
    {
        'key': 'MENU_BG',
        'type': 'color',
        'display_name': _('Menu/toolbar background')
    },
    {
        'key': 'MENU_FG',
        'type': 'color',
        'display_name': _('Menu/toolbar text'),
    },
    {
        'key': 'SEL_BG',
        'type': 'color',
        'display_name': _('Selection highlight')
    },
    {
        'key': 'SEL_FG',
        'type': 'color',
        'display_name': _('Selection text'),
    },
    {
        'key': 'ACCENT_BG',
        'fallback_key': 'SEL_BG',
        'type': 'color',
        'display_name': _('Accent color (checkboxes, radios)'),
    },
    {
        'key': 'TXT_BG',
        'type': 'color',
        'display_name': _('Textbox background')
    },
    {
        'key': 'TXT_FG',
        'type': 'color',
        'display_name': _('Textbox text'),
    },
    {
        'key': 'BTN_BG',
        'type': 'color',
        'display_name': _('Button background')
    },
    {
        'key': 'BTN_FG',
        'type': 'color',
        'display_name': _('Button text'),
    },
    {
        'key': 'HDR_BTN_BG',
        'fallback_key': 'BTN_BG',
        'type': 'color',
        'display_name': _('Header button background'),
    },
    {
        'key': 'HDR_BTN_FG',
        'fallback_key': 'BTN_FG',
        'type': 'color',
        'display_name': _('Header button text'),
    },
    {
        'key': 'WM_BORDER_FOCUS',
        'fallback_key': 'SEL_BG',
        'type': 'color',
        'display_name': _('Focused window border'),
    },
    {
        'key': 'WM_BORDER_UNFOCUS',
        'fallback_key': 'MENU_BG',
        'type': 'color',
        'display_name': _('Unfocused window border'),
    },
]
merge_theme_model_with_base(THEME_MODEL, BASE_THEME_MODEL_GTK, 'gtk')

BASE_THEME_MODEL_OPTIONS = [
    {
        'type': 'separator',
        'display_name': _('Theme options'),
    },
    {
        'key': 'ROUNDNESS',
        'type': 'int',
        'fallback_value': 2,
        'display_name': _('Roundness'),
    },
    {
        'key': 'SPACING',
        'type': 'int',
        'fallback_value': 3,
        'display_name': _('(GTK3) Spacing'),
    },
    {
        'key': 'GRADIENT',
        'type': 'float',
        'fallback_value': 0.0,
        'display_name': _('(GTK3) Gradient'),
    },
    {
        'key': 'GTK3_GENERATE_DARK',
        'type': 'bool',
        'fallback_value': True,
        'display_name': _('(GTK3) Add dark variant'),
    },
]
merge_theme_model_with_base(THEME_MODEL, BASE_THEME_MODEL_OPTIONS, 'options')

BASE_ICON_THEME_MODEL = [
    {
        'type': 'separator',
        'display_name': _('Iconset')
    },
    {
        'key': 'ICONS_STYLE',
        'type': 'options',
        'options': [
            {
                'value': icons_plugin.name,
                'display_name': icons_plugin.display_name,
            }
            for icons_plugin in ICONS_PLUGINS.values()
        ],
        'fallback_value': 'gnome_colors',
        'display_name': _('Icons style')
    },
]
THEME_MODEL += BASE_ICON_THEME_MODEL
merge_model_with_base(
    whole_theme_model=THEME_MODEL,
    plugin_model_name='icons',
    value_filter_key='ICONS_STYLE',
    plugins=ICONS_PLUGINS,
)

THEME_MODEL += [
    {
        'type': 'separator',
        'display_name': _('Terminal')
    },
    {
        'key': 'TERMINAL_THEME_MODE',
        'type': 'options',
        'options': [
            {'value': 'auto', 'display_name': _('Auto')},
            {'value': 'basic', 'display_name': _('Basic')},
            {'value': 'manual', 'display_name': _('Manual')},
            {'value': 'smarty', 'display_name': _('Experimental')},
        ],
        'fallback_value': 'auto',
        'display_name': _('Theme options')
    },
    {
        'key': 'TERMINAL_BASE_TEMPLATE',
        'type': 'options',
        'options': [
            {'value': template_name}
            for template_name in sorted(os.listdir(TERMINAL_TEMPLATE_DIR))
        ],
        'fallback_value': 'monovedek',
        'display_name': _('Theme style'),
        'value_filter': {
            'TERMINAL_THEME_MODE': ['auto', 'basic', 'smarty', ],
        },
    },
    {
        'key': 'TERMINAL_THEME_AUTO_BGFG',
        'type': 'bool',
        'fallback_value': True,
        'display_name': _('Auto-swap BG/FG'),
        'value_filter': {
            'TERMINAL_THEME_MODE': ['auto', 'basic', 'smarty', ],
        },
    },
    {
        'key': 'TERMINAL_THEME_EXTEND_PALETTE',
        'type': 'bool',
        'fallback_value': False,
        'display_name': _('Extend palette with more lighter/darker colors'),
        'value_filter': {
            'TERMINAL_THEME_MODE': ['smarty', ],
        },
    },
    {
        'key': 'TERMINAL_THEME_ACCURACY',
        'type': 'int',
        'fallback_value': 128,
        'display_name': _('Palette generation accuracy'),
        'value_filter': {
            'TERMINAL_THEME_MODE': ['smarty', ],
        },
        'min_value': 8,
        'max_value': 255,
    },
    {
        'key': 'TERMINAL_BACKGROUND',
        'type': 'color',
        'fallback_key': 'TXT_BG',
        # 'fallback_key': 'MENU_BG',
        'display_name': _('Terminal background'),
        'value_filter': {
            'TERMINAL_THEME_MODE': ['basic', 'manual'],
        },
    },
    {
        'key': 'TERMINAL_FOREGROUND',
        'type': 'color',
        'fallback_key': 'TXT_FG',
        # 'fallback_key': 'MENU_FG',
        'display_name': _('Terminal foreground'),
        'value_filter': {
            'TERMINAL_THEME_MODE': ['basic', 'manual'],
        },
    },
    {
        'key': 'TERMINAL_ACCENT_COLOR',
        'type': 'color',
        'fallback_key': 'SEL_BG',
        'display_name': _('Terminal accent color'),
        'value_filter': {
            'TERMINAL_THEME_MODE': ['basic', ],
        },
    },
    # Black
    {
        'key': 'TERMINAL_COLOR0',
        'type': 'color',
        'display_name': _('Black'),
        'value_filter': {
            'TERMINAL_THEME_MODE': ['manual', ],
        },
    },
    {
        'key': 'TERMINAL_COLOR8',
        'type': 'color',
        'display_name': _('Black highlight'),
        'value_filter': {
            'TERMINAL_THEME_MODE': ['manual', ],
        },
    },
    # Red
    {
        'key': 'TERMINAL_COLOR1',
        'type': 'color',
        'display_name': _('Red'),
        'value_filter': {
            'TERMINAL_THEME_MODE': ['manual', ],
        },
    },
    {
        'key': 'TERMINAL_COLOR9',
        'type': 'color',
        'display_name': _('Red highlight'),
        'value_filter': {
            'TERMINAL_THEME_MODE': ['manual', ],
        },
    },
    # Green
    {
        'key': 'TERMINAL_COLOR2',
        'type': 'color',
        'display_name': _('Green'),
        'value_filter': {
            'TERMINAL_THEME_MODE': ['manual', ],
        },
    },
    {
        'key': 'TERMINAL_COLOR10',
        'type': 'color',
        'display_name': _('Green highlight'),
        'value_filter': {
            'TERMINAL_THEME_MODE': ['manual', ],
        },
    },
    # Yellow
    {
        'key': 'TERMINAL_COLOR3',
        'type': 'color',
        'display_name': _('Yellow'),
        'value_filter': {
            'TERMINAL_THEME_MODE': ['manual', ],
        },
    },
    {
        'key': 'TERMINAL_COLOR11',
        'type': 'color',
        'display_name': _('Yellow highlight'),
        'value_filter': {
            'TERMINAL_THEME_MODE': ['manual', ],
        },
    },
    # Blue
    {
        'key': 'TERMINAL_COLOR4',
        'type': 'color',
        'display_name': _('Blue'),
        'value_filter': {
            'TERMINAL_THEME_MODE': ['manual', ],
        },
    },
    {
        'key': 'TERMINAL_COLOR12',
        'type': 'color',
        'display_name': _('Blue highlight'),
        'value_filter': {
            'TERMINAL_THEME_MODE': ['manual', ],
        },
    },
    # Purple
    {
        'key': 'TERMINAL_COLOR5',
        'type': 'color',
        'display_name': _('Purple'),
        'value_filter': {
            'TERMINAL_THEME_MODE': ['manual', ],
        },
    },
    {
        'key': 'TERMINAL_COLOR13',
        'type': 'color',
        'display_name': _('Purple highlight'),
        'value_filter': {
            'TERMINAL_THEME_MODE': ['manual', ],
        },
    },
    # Cyan
    {
        'key': 'TERMINAL_COLOR6',
        'type': 'color',
        'display_name': _('Cyan'),
        'value_filter': {
            'TERMINAL_THEME_MODE': ['manual', ],
        },
    },
    {
        'key': 'TERMINAL_COLOR14',
        'type': 'color',
        'display_name': _('Cyan highlight'),
        'value_filter': {
            'TERMINAL_THEME_MODE': ['manual', ],
        },
    },
    # White
    {
        'key': 'TERMINAL_COLOR7',
        'type': 'color',
        'display_name': _('White'),
        'value_filter': {
            'TERMINAL_THEME_MODE': ['manual', ],
        },
    },
    {
        'key': 'TERMINAL_COLOR15',
        'type': 'color',
        'display_name': _('White highlight'),
        'value_filter': {
            'TERMINAL_THEME_MODE': ['manual', ],
        },
    },
]

merge_theme_model_with_base(THEME_MODEL, [], 'extra')
merge_model_with_base(
    whole_theme_model=THEME_MODEL,
    plugin_model_name='extra',
    plugins=EXPORT_PLUGINS,
)

BASE_THEME_MODEL_OTHER = [
    {
        'type': 'separator',
        'display_name': _('Other options'),
    },
]
merge_theme_model_with_base(THEME_MODEL, BASE_THEME_MODEL_OTHER, 'other')

THEME_MODEL_BY_KEY = {
    value['key']: value
    for value in THEME_MODEL
    if 'key' in value
}



import os
import sys
import re

from .config import TERMINAL_TEMPLATE_DIR
from .color import (
    SMALLEST_DIFF, ColorDiff, is_dark,
    hex_to_int, color_list_from_hex, color_hex_from_list, hex_darker
)


RED = 0
GREEN = 1
BLUE = 2


def find_closest_color_key(color_hex, colors_hex, highlight=True):
    smallest_diff = SMALLEST_DIFF
    smallest_key = None
    highlight_keys = ["color{}".format(i) for i in range(8, 15 + 1)]
    for preset_key, preset_color in colors_hex.items():
        if (
                highlight
        ) and (
            "color" not in preset_key
        ) or (
            preset_key not in highlight_keys
        ):
            continue
        diff = ColorDiff(preset_color, color_hex)
        # if diff.minabs < smallest_diff.minabs:
        if diff.abs < smallest_diff.abs:
            smallest_diff = diff
            smallest_key = preset_key
    return smallest_key, smallest_diff


VALID_COLOR_CHARS = [
    chr(i) for i in range(ord('a'), ord('f') + 1)
] + [
    str(i) for i in range(10)
]


def import_xcolors(path):
    hex_colors = {}
    with open(os.path.expanduser(path)) as file_object:
        for line in file_object.read().split('\n'):
            if line.strip().startswith('!'):
                continue
            pair = list([s.strip() for s in line.split(':')])
            if len(pair) < 2:
                continue
            key, value = pair
            key = key.replace('*', '')
            value = value.replace('#', '').lower()
            for char in value:
                if char not in VALID_COLOR_CHARS:
                    break
            else:
                hex_colors[key] = value
    return hex_colors


def generate_theme_from_hint(  # pylint: disable=too-many-arguments
        template_path, theme_color, theme_bg, theme_fg,
        theme_hint=None, auto_swap_colors=True
):
    hex_colors = import_xcolors(template_path)
    if auto_swap_colors and (
            is_dark(theme_bg) != is_dark(hex_colors['background'])
    ):
        theme_bg, theme_fg = theme_fg, theme_bg
    _closest_key, diff = None, None
    if theme_hint:
        _closest_key = theme_hint
        diff = ColorDiff(hex_colors[theme_hint], theme_color)
    else:
        _closest_key, diff = find_closest_color_key(
            theme_color, hex_colors, highlight=False
        )
    modified_colors = {
        key: diff.apply_to(value)
        for key, value in hex_colors.items()
    }
    modified_colors["background"] = theme_bg
    modified_colors["foreground"] = theme_fg
    for source, destinations in {
            "background": ("color0", "color8",),
            "foreground": ("color7", "color15",),
    }.items():
        for key in destinations:
            modified_colors[key] = ColorDiff(
                hex_colors[source], hex_colors[key]
            ).apply_to(
                modified_colors[source]
            )
    return modified_colors


def get_bright_colors(palette, brightness_margin=20):
    # brightness_margin = 30
    # brightness_margin = 40
    list_of_colors = [[hex_to_int(s) for s in color_list_from_hex(c)] for c in palette]
    bright_colors = []
    for color in list_of_colors:
        if max(
                abs(color[RED]-color[GREEN]) > brightness_margin,
                abs(color[RED]-color[BLUE]) > brightness_margin,
                abs(color[GREEN]-color[RED]) > brightness_margin,
                abs(color[GREEN]-color[BLUE]) > brightness_margin,
                abs(color[BLUE]-color[GREEN]) > brightness_margin,
                abs(color[BLUE]-color[RED]) > brightness_margin
        ):
            bright_colors.append(color)

    bright_colors.sort(key=sum)
    bright_colors = [color_hex_from_list(c) for c in bright_colors if 565 > sum(c) > 200]
    return bright_colors


def get_all_colors_from_oomox_colorscheme(palette):
    from .theme_model import THEME_MODEL

    all_colors = []
    for theme_model_item in THEME_MODEL:
        if theme_model_item.get('type') != 'color':
            continue
        color_name = theme_model_item.get('key')
        # if not color_name or color_name.startswith('TERMINAL') or color_name.startswith('ICON'):
        if not color_name or color_name.startswith('TERMINAL'):
            continue
        color_value = palette.get(theme_model_item['key'])
        if not color_value or color_value in all_colors:
            continue
        all_colors.append(color_value)
    return all_colors


class ContinueNext(Exception):
    pass


# @TODO:
# These two functions are temporary until progressbar API won't be implemented in UI
def get_term_width():
    import shutil
    return shutil.get_terminal_size((80, 80)).columns


class ProgressBar():

    message = None
    print_ratio = None
    index = 0
    progress = 0

    LEFT_DECORATION = '['
    RIGHT_DECORATION = ']'
    EMPTY = '-'
    FULL = '#'

    def __init__(self, length, message=None):
        message = message or str(length)
        self.message = message
        width = (
            get_term_width() - len(message) -
            len(self.LEFT_DECORATION) - len(self.RIGHT_DECORATION)
        )
        self.print_ratio = length / width
        sys.stderr.write(message)
        sys.stderr.write(self.LEFT_DECORATION + self.EMPTY * width + self.RIGHT_DECORATION)
        sys.stderr.write('{}[\bb'.format(chr(27)) * (width + len(self.RIGHT_DECORATION)))
        sys.stderr.flush()

    def update(self):
        self.index += 1
        if self.index / self.print_ratio > self.progress:
            self.progress += 1
            sys.stderr.write(self.FULL)
            sys.stderr.flush()

    def __enter__(self):
        return self.update

    def __exit__(self, *exc_details):
        sys.stderr.write('\n')
# ######## END


def _generate_theme_from_full_palette(
        template_path, all_colors, accuracy=None, extend_palette=False
):  # pylint: disable=invalid-name,too-many-nested-blocks,too-many-locals,too-many-statements,too-many-branches
    # @TODO: refactor it some day :3

    # how far should be the colors to be counted as similar (0 .. 255*3)
    # DIFF_MARGIN = 30
    DIFF_MARGIN = 60

    # criterias to recognize bright colors (0 .. 255*3)
    MIN_LIGHTNESS = 10
    BRIGHTNESS_MARGIN = 20

    # 1 means similarity to template the same important as mathing color palette
    # SIMILARITY_IMPORTANCE = 2
    SIMILARITY_IMPORTANCE = 2.5

    accuracy = accuracy or 0x20
    hex_colors = import_xcolors(template_path)
    hex_colors_as_color_lists = {
        key: [
            hex_to_int(s) for s in color_list_from_hex(value)
        ] for key, value in hex_colors.items()
    }
    if extend_palette:
        for color in all_colors[:]:
            for i in (20, 40, 60):
                all_colors.append(hex_darker(color, i))
                all_colors.append(hex_darker(color, -i))

    bright_colors = get_bright_colors(all_colors, brightness_margin=BRIGHTNESS_MARGIN)
    bright_colors_as_color_lists = [
        [
            hex_to_int(s) for s in color_list_from_hex(value)
        ] for value in bright_colors
    ]

    START = [-0xff, -0xff, -0xff]
    END = [0xff, 0xff, 0xff]
    best_diff_color_values = [0, 0, 0]
    biggest_number_of_similar = None
    prev_biggest_number_of_similar = 'not_found'
    best_result = None

    _debug_iteration_counter = 0

    while accuracy > 0:
        _debug_iteration_counter += 1
        print()
        print(('ITERATION', _debug_iteration_counter))
        progress = ProgressBar(
            length=((int(abs(START[0] - END[0])/accuracy) + 2) ** 3)
        )
        red = START[RED]
        while red < END[RED] + accuracy:
            green = START[GREEN]
            while green < END[GREEN] + accuracy:
                blue = START[BLUE]
                while blue < END[BLUE] + accuracy:
                    try:

                        color_list = [red, green, blue]
                        modified_colors = {}
                        for key, value in hex_colors_as_color_lists.items():
                            if not key.startswith('color'):
                                continue
                            new_value = value[:]
                            for i in range(3):
                                new_value[i] = min(
                                    255,
                                    max(
                                        0,
                                        new_value[i] + (red, green, blue)[i]
                                    )
                                )
                            if key not in ['color0', 'color7', 'color8', 'color15']:
                                if (
                                        MIN_LIGHTNESS > sum(new_value)
                                ) or (
                                    sum(new_value) > (255*3 - MIN_LIGHTNESS)
                                ):
                                    raise ContinueNext()
                            modified_colors[key] = new_value

                        num_of_similar = 0
                        for modified_color in modified_colors.values():
                            for bright_color in bright_colors_as_color_lists:
                                abs_diff = 0
                                for i in range(3):
                                    abs_diff += abs(modified_color[i] - bright_color[i])
                                if abs_diff < DIFF_MARGIN:
                                    num_of_similar += 1

                        similarity_to_reference = (
                            255*3 - sum([abs(c) for c in color_list]) * SIMILARITY_IMPORTANCE
                        ) / (255*3)
                        num_of_similar *= similarity_to_reference

                        if (
                                biggest_number_of_similar is None
                        ) or (
                            num_of_similar > biggest_number_of_similar
                        ):
                            biggest_number_of_similar = num_of_similar
                            best_result = modified_colors
                            best_diff_color_values[RED] = red
                            best_diff_color_values[GREEN] = green
                            best_diff_color_values[BLUE] = blue

                    except ContinueNext:
                        pass
                    progress.update()
                    blue += accuracy
                green += accuracy
            red += accuracy

        if biggest_number_of_similar == prev_biggest_number_of_similar:
            print('good enough')
            break
        prev_biggest_number_of_similar = biggest_number_of_similar
        for i in range(3):
            START[i] = max(best_diff_color_values[i] - accuracy, -255)
            END[i] = min(best_diff_color_values[i] + accuracy, 255)
        accuracy = round(accuracy / 2)
        print(('DEEPER!', accuracy))

    # from fabulous.color import bg256
    # for bright_color in bright_colors:
    #     print(bg256(bright_color, bright_color))

    modified_colors = {
        key: color_hex_from_list(c)
        for key, c in best_result.items()
    }
    return modified_colors


_FULL_PALETTE_CACHE = {}


def generate_theme_from_full_palette(
        palette, theme_bg, theme_fg, *args, auto_swap_colors=True, **kwargs
):  # pylint: disable=invalid-name
    all_colors = sorted(get_all_colors_from_oomox_colorscheme(palette))
    cache_id = str(
        list(args) + [
            kwargs[name] for name in sorted(kwargs, key=lambda x: x[0])
        ] + all_colors
    )
    if cache_id not in _FULL_PALETTE_CACHE:
        # from time import time
        # before = time()
        _FULL_PALETTE_CACHE[cache_id] = _generate_theme_from_full_palette(
            *args, all_colors=all_colors, **kwargs
        )
        # print(time() - before)
    modified_colors = {}
    modified_colors.update(_FULL_PALETTE_CACHE[cache_id])
    if auto_swap_colors:
        theme_bg, theme_fg = theme_fg, theme_bg
    modified_colors["background"] = theme_bg
    modified_colors["foreground"] = theme_fg
    return modified_colors


def generate_themes_from_oomox(original_colorscheme):
    colorscheme = {}
    colorscheme.update(original_colorscheme)
    term_colorscheme = None
    if colorscheme['TERMINAL_THEME_MODE'] in ('auto', 'smarty'):
        colorscheme["TERMINAL_ACCENT_COLOR"] = colorscheme["SEL_BG"]
        colorscheme["TERMINAL_BACKGROUND"] = colorscheme["TXT_BG"]
        colorscheme["TERMINAL_FOREGROUND"] = colorscheme["TXT_FG"]
        if colorscheme["THEME_STYLE"] == "materia":
            colorscheme["TERMINAL_FOREGROUND"] = colorscheme["FG"]
    if colorscheme['TERMINAL_THEME_MODE'] == 'smarty':
        term_colorscheme = generate_theme_from_full_palette(
            template_path=os.path.join(
                TERMINAL_TEMPLATE_DIR, colorscheme["TERMINAL_BASE_TEMPLATE"]
            ),
            palette=colorscheme,
            theme_bg=colorscheme["TERMINAL_BACKGROUND"],
            theme_fg=colorscheme["TERMINAL_FOREGROUND"],
            auto_swap_colors=colorscheme["TERMINAL_THEME_AUTO_BGFG"],
            extend_palette=colorscheme["TERMINAL_THEME_EXTEND_PALETTE"],
            accuracy=colorscheme.get("TERMINAL_THEME_ACCURACY")
        )
    elif colorscheme['TERMINAL_THEME_MODE'] in ('basic', 'auto'):
        term_colorscheme = generate_theme_from_hint(
            template_path=os.path.join(
                TERMINAL_TEMPLATE_DIR, colorscheme["TERMINAL_BASE_TEMPLATE"]
            ),
            theme_color=colorscheme["TERMINAL_ACCENT_COLOR"],
            theme_bg=colorscheme["TERMINAL_BACKGROUND"],
            theme_fg=colorscheme["TERMINAL_FOREGROUND"],
            theme_hint=None,
            auto_swap_colors=colorscheme["TERMINAL_THEME_AUTO_BGFG"]
        )
    else:
        term_colorscheme = convert_oomox_theme_to_xrdb(colorscheme)
    for i in range(16):
        theme_key = "TERMINAL_COLOR{}".format(i)
        term_key = "color{}".format(i)
        colorscheme[theme_key] = term_colorscheme[term_key]
    if colorscheme['TERMINAL_THEME_MODE'] != 'manual':
        colorscheme['TERMINAL_BACKGROUND'] = term_colorscheme['background']
        colorscheme['TERMINAL_FOREGROUND'] = term_colorscheme['foreground']
    return term_colorscheme, colorscheme


def convert_oomox_theme_to_xrdb(colorscheme):
    term_colorscheme = {}
    for i in range(16):
        theme_key = "TERMINAL_COLOR{}".format(i)
        term_key = "color{}".format(i)
        if colorscheme.get(theme_key):
            term_colorscheme[term_key] = \
                    colorscheme[theme_key]
    term_colorscheme['background'] = colorscheme['TERMINAL_BACKGROUND']
    term_colorscheme['foreground'] = colorscheme['TERMINAL_FOREGROUND']
    return term_colorscheme


def generate_xrdb_theme_from_oomox(colorscheme):
    return convert_oomox_theme_to_xrdb(colorscheme)


def generate_terminal_colors_for_oomox(colorscheme):
    _, new_oomox_colorscheme = generate_themes_from_oomox(colorscheme)
    return new_oomox_colorscheme


def natural_sort(list_to_sort):
    def convert(text):
        return int(text) if text.isdigit() else text.lower()

    def alphanum_key(key):
        return [convert(c) for c in re.split('([0-9]+)', key)]

    return sorted(list_to_sort, key=alphanum_key)


def generate_xresources(colorscheme):
    color_keys = colorscheme.keys()
    color_regex = re.compile('color[0-9]')
    return '\n'.join([
        "*{key}:  #{value}".format(
            key=key, value=colorscheme[key]
        )
        for key in (
            sorted([
                key for key in color_keys
                if not color_regex.match(key)
            ]) +
            natural_sort([
                key for key in color_keys
                if color_regex.match(key)
            ])
        )
    ])


def cli():
    args = sys.argv
    if len(args) < 5:
        print(
            "Usage: {} "
            "TEMPLATE_PATH ACCENT_COLOR BG FG "
            "[ACCENT_KEY_NAME] [AUTO_DETECT_FG_BG=YES]".format(
                sys.argv[0]
            )
        )
        sys.exit(1)
    template_path = args[1]
    theme_color = args[2]
    theme_bg = args[3]
    theme_fg = args[4]
    theme_hint = args[5] if len(args) > 5 else None
    auto_swap_colors = (args[6] not in ["y", "yes", "true", "1"]) \
        if len(args) > 6 else None
    term_colorscheme = generate_theme_from_hint(
        template_path=template_path,
        theme_color=theme_color,
        theme_bg=theme_bg,
        theme_fg=theme_fg,
        theme_hint=theme_hint,
        auto_swap_colors=auto_swap_colors,
    )
    print(generate_xresources(term_colorscheme))


if __name__ == '__main__':
    cli()



import json

from .config import USER_PALETTE_PATH


class PaletteCache():

    _palette_cache = None

    @staticmethod
    def load():
        try:
            with open(USER_PALETTE_PATH, 'r') as file_object:
                return json.load(file_object)
        except FileNotFoundError:
            return []

    @staticmethod
    def save(palette_cache_list):
        with open(USER_PALETTE_PATH, 'w') as file_object:
            return json.dump(palette_cache_list, file_object)

    @classmethod
    def set(cls, palette_cache_list):
        cls._palette_cache = palette_cache_list
        cls.save(palette_cache_list)

    @classmethod
    def get(cls):
        if not cls._palette_cache:
            cls._palette_cache = cls.load()
        return cls._palette_cache

    @classmethod
    def get_gtk(cls):
        return ':'.join(cls.get())

    @classmethod
    def add_color(cls, gtk_color):
        gtk_color_converted = gtk_color.to_color().to_string()
        palette_cache_list = [
            string for string in cls.get()  # pylint: disable=not-an-iterable
            if string != ''
        ]
        if gtk_color_converted not in palette_cache_list:
            cls.set(
                ([gtk_color_converted] + palette_cache_list)[:20]
            )



import os

from oomox_gui.config import FALLBACK_COLOR
from oomox_gui.export_common import FileBasedExportDialog
from oomox_gui.plugin_api import OomoxIconsPlugin
from oomox_gui.i18n import _


PLUGIN_DIR = os.path.dirname(os.path.realpath(__file__))
GNOME_COLORS_ICON_THEME_DIR = os.path.join(PLUGIN_DIR, "gnome-colors-icon-theme/")


class GnomeColorsIconsExportDialog(FileBasedExportDialog):
    timeout = 600

    def do_export(self):
        self.command = [
            "bash",
            os.path.join(GNOME_COLORS_ICON_THEME_DIR, "change_color.sh"),
            "-o", self.theme_name,
            self.temp_theme_path,
        ]
        super().do_export()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.do_export()


class Plugin(OomoxIconsPlugin):

    name = 'gnome_colors'
    display_name = 'Gnome-Colors'
    export_dialog = GnomeColorsIconsExportDialog
    preview_svg_dir = os.path.join(PLUGIN_DIR, "icon_previews/")

    theme_model_icons = [
        {
            'key': 'ICONS_LIGHT_FOLDER',
            'type': 'color',
            'fallback_key': 'SEL_BG',
            'display_name': _('Light base (folders)'),
        },
        {
            'key': 'ICONS_LIGHT',
            'fallback_key': 'SEL_BG',
            'type': 'color',
            'display_name': _('Light base'),
        },
        {
            'key': 'ICONS_MEDIUM',
            'type': 'color',
            'fallback_key': 'BTN_BG',
            'display_name': _('Medium base'),
        },
        {
            'key': 'ICONS_DARK',
            'type': 'color',
            'fallback_key': 'MENU_BG',
            'display_name': _('Dark stroke'),
        },
    ]

    def preview_transform_function(self, svg_template, colorscheme):
        return svg_template.replace(
            "LightFolderBase", colorscheme["ICONS_LIGHT_FOLDER"] or FALLBACK_COLOR
        ).replace(
            "LightBase", colorscheme["ICONS_LIGHT"] or FALLBACK_COLOR
        ).replace(
            "MediumBase", colorscheme["ICONS_MEDIUM"] or FALLBACK_COLOR
        ).replace(
            "DarkStroke", colorscheme["ICONS_DARK"] or FALLBACK_COLOR
        )



# pylint: disable=too-few-public-methods
import os

from gi.repository import Gtk

from oomox_gui.export_common import GtkThemeExportDialog
# from oomox_gui.export_common import OPTION_GTK2_HIDPI
from oomox_gui.plugin_api import OomoxThemePlugin
from oomox_gui.color import mix_theme_colors
from oomox_gui.i18n import _


PLUGIN_DIR = os.path.dirname(os.path.realpath(__file__))
THEME_DIR = os.path.join(PLUGIN_DIR, "arc-theme/")

OPTION_EXPORT_CINNAMON_THEME = 'OPTION_EXPORT_CINNAMON_THEME'
OPTION_EXPORT_GNOME_SHELL_THEME = 'OPTION_EXPORT_GNOME_SHELL_THEME'
OPTION_EXPORT_XFWM_THEME = 'OPTION_EXPORT_XFWM_THEME'


class ArcThemeExportDialog(GtkThemeExportDialog):

    config_name = 'arc_theme'
    timeout = 1000

    def do_export(self):
        self.command = [
            "bash",
            os.path.join(THEME_DIR, "change_color.sh"),
            # "--hidpi", str(self.export_config[OPTION_GTK2_HIDPI]),
            "--output", self.theme_name,
            self.temp_theme_path,
        ]
        autogen_opts = []
        if not self.export_config[OPTION_EXPORT_CINNAMON_THEME]:
            autogen_opts += ["--disable-cinnamon"]
        if not self.export_config[OPTION_EXPORT_GNOME_SHELL_THEME]:
            autogen_opts += ["--disable-gnome-shell"]
        if not self.export_config[OPTION_EXPORT_XFWM_THEME]:
            autogen_opts += ["--disable-xfwm"]
        if autogen_opts:
            self.command += [
                "--autogen-opts", " ".join(autogen_opts),
            ]
        super().do_export()

    def __init__(self, transient_for, colorscheme, theme_name, **kwargs):
        super().__init__(
            transient_for=transient_for,
            colorscheme=colorscheme,
            theme_name=theme_name,
            override_options={
                OPTION_EXPORT_CINNAMON_THEME: {
                    'default': False,
                    'display_name': _("Generate theme for _Cinnamon"),
                },
                OPTION_EXPORT_GNOME_SHELL_THEME: {
                    'default': False,
                    'display_name': _("Generate theme for Gnome _Shell"),
                },
                OPTION_EXPORT_XFWM_THEME: {
                    'default': False,
                    'display_name': _("Generate theme for _Xfwm"),
                },
            },
            **kwargs
        )


def _monkeypatch_update_preview_borders(preview_object):
    _monkeypatch_id = '_arc_borders_monkeypatched'

    if getattr(preview_object, _monkeypatch_id, None):
        return

    old_update_preview_borders = preview_object.update_preview_borders

    def _update_preview_borders(colorscheme):
        if colorscheme["THEME_STYLE"] != "arc":
            old_update_preview_borders(colorscheme)
        else:
            for widget_name, widget, border_color in (  # pylint: disable=invalid-name
                    (
                        'button',
                        preview_object.gtk_preview.button,
                        colorscheme['ARC_WIDGET_BORDER_COLOR'],
                    ), (
                        'headerbar_button',
                        preview_object.gtk_preview.headerbar.button,
                        mix_theme_colors(
                            colorscheme['HDR_BTN_FG'],
                            colorscheme['HDR_BTN_BG'],
                            0.12
                        ),
                    ), (
                        'entry',
                        preview_object.gtk_preview.entry,
                        colorscheme['ARC_WIDGET_BORDER_COLOR'],
                    ),
            ):
                css_provider_border_color = preview_object.css_providers.border.get(widget_name)
                if not css_provider_border_color:
                    css_provider_border_color = \
                        preview_object.css_providers.border[widget_name] = \
                        Gtk.CssProvider()
                css_provider_border_color.load_from_data(
                    """
                    * {{
                        border-color: #{border_color};
                        border-radius: {roundness}px;
                    }}
                    """.format(
                        border_color=border_color,
                        roundness=colorscheme["ROUNDNESS"],
                    ).encode('ascii')
                )
                Gtk.StyleContext.add_provider(
                    widget.get_style_context(),
                    css_provider_border_color,
                    Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION
                )

    preview_object.update_preview_borders = _update_preview_borders
    setattr(preview_object, _monkeypatch_id, True)


class Plugin(OomoxThemePlugin):

    name = 'arc'
    display_name = 'Arc'
    description = '(GTK+2, GTK+3, Cinnamon, Gnome Shell, Metacity, Openbox, Unity, Xfwm)'
    export_dialog = ArcThemeExportDialog
    gtk_preview_dir = os.path.join(PLUGIN_DIR, "gtk_preview_css/")
    preview_sizes = {
        OomoxThemePlugin.PreviewImageboxesNames.CHECKBOX.name: 24,
    }

    enabled_keys_gtk = [
        'BG',
        'FG',
        'MENU_BG',
        'MENU_FG',
        'SEL_BG',
        'SEL_FG',
        'TXT_BG',
        'BTN_BG',
        'HDR_BTN_BG',
        'ACCENT_BG',
    ]
    # enabled_keys_options = [
    #     'ROUNDNESS',
    # ]

    theme_model_gtk = [
        {
            'key': 'ARC_WIDGET_BORDER_COLOR',
            'fallback_function': lambda colors: mix_theme_colors(
                colors['BTN_BG'], colors['BTN_FG'],
                0.75
            ),
            'type': 'color',
            'display_name': _('(GTK3) Widget border color'),
        },
    ]

    theme_model_options = [
        {
            'key': 'ARC_TRANSPARENCY',
            'type': 'bool',
            'fallback_value': True,
            'display_name': _('(GTK3) Enable theme transparency'),
        },
        # {
        #     'key': 'GTK3_GENERATE_DARK',
        #     'type': 'bool',
        #     'fallback_value': True,
        #     'display_name': _('(GTK3) Add dark variant'),
        # },
    ]

    """
    theme_model_other = [
        {
            'key': 'GNOME_SHELL_PANEL_OPACITY',
            'type': 'float',
            'fallback_value': 0.6,
            'max_value': 1.0,
            'display_name': _('(Gnome Shell) Panel opacity'),
        },
    ]
    """

    def preview_before_load_callback(self, preview_object, colorscheme):
        colorscheme["TXT_FG"] = colorscheme["FG"]
        colorscheme["BTN_FG"] = colorscheme["FG"]
        colorscheme["HDR_BTN_FG"] = colorscheme["MENU_FG"]
        colorscheme["WM_BORDER_FOCUS"] = colorscheme["MENU_BG"]
        colorscheme["WM_BORDER_UNFOCUS"] = colorscheme["BTN_BG"]
        colorscheme["GRADIENT"] = 0
        colorscheme["ROUNDNESS"] = 0
        preview_object.WM_BORDER_WIDTH = 0
        _monkeypatch_update_preview_borders(preview_object)



import os

from oomox_gui.config import FALLBACK_COLOR
from oomox_gui.export_common import FileBasedExportDialog
from oomox_gui.plugin_api import OomoxIconsPlugin
from oomox_gui.i18n import _


PLUGIN_DIR = os.path.dirname(os.path.realpath(__file__))
ARCHDROID_THEME_DIR = os.path.join(PLUGIN_DIR, "archdroid-icon-theme/")


class ArchdroidIconsExportDialog(FileBasedExportDialog):
    timeout = 100

    def do_export(self):
        self.command = [
            "bash",
            os.path.join(ARCHDROID_THEME_DIR, "change_color.sh"),
            "-o", self.theme_name,
            self.temp_theme_path,
        ]
        super().do_export()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.do_export()


class Plugin(OomoxIconsPlugin):

    name = 'archdroid'
    display_name = 'Archdroid'
    export_dialog = ArchdroidIconsExportDialog
    preview_svg_dir = os.path.join(PLUGIN_DIR, "icon_previews/")

    theme_model_icons = [
        {
            'key': 'ICONS_ARCHDROID',
            'type': 'color',
            'fallback_key': 'SEL_BG',
            'display_name': _('Icons color'),
        },
    ]

    def preview_transform_function(self, svg_template, colorscheme):
        return svg_template.replace(
            "%ICONS_ARCHDROID%", colorscheme["ICONS_ARCHDROID"] or FALLBACK_COLOR
        )



import os
import json

from .config import USER_EXPORT_CONFIG_DIR


class ExportConfig():

    name = None
    config_path = None
    config = None
    default_config = None

    def __init__(self, config_name, default_config=None):
        self.name = config_name
        self.default_config = default_config
        self.config = self.default_config or {}
        self.config_path = os.path.join(
            USER_EXPORT_CONFIG_DIR,
            "{}.json".format(self.name)
        )
        self.load()

    def load(self):
        try:
            with open(self.config_path, 'r') as file_object:
                for key, value in json.load(file_object).items():
                    self.config[key] = value
        except FileNotFoundError:
            pass
        return self.config

    def save(self):
        if not os.path.exists(USER_EXPORT_CONFIG_DIR):
            os.makedirs(USER_EXPORT_CONFIG_DIR)
        with open(self.config_path, 'w') as file_object:
            return json.dump(self.config, file_object)

    def __getitem__(self, item):
        return self.config[item]

    def __setitem__(self, item, value):
        self.config[item] = value

    def __str__(self):
        return str(self.config)



import os
import gc

from oomox_gui.plugin_api import OomoxImportPlugin
from oomox_gui.config import TERMINAL_TEMPLATE_DIR
from oomox_gui.color import (
    hex_to_int, color_list_from_hex, color_hex_from_list,
    find_closest_color, hex_darker
)
from oomox_gui.terminal import (
    import_xcolors,
)
from oomox_gui.helpers import get_plugin_module
from oomox_gui.i18n import _


PLUGIN_DIR = os.path.dirname(os.path.realpath(__file__))


LOW_QUALITY = 100
MEDIUM_QUALITY = 200
HIGH_QUALITY = 400
ULTRA_QUALITY = 1000


image_analyzer = get_plugin_module('ima', os.path.join(PLUGIN_DIR, 'ima.py'))  # pylint: disable=invalid-name


def sort_by_saturation(c):
    # pylint: disable=invalid-name
    return abs(c[0]-c[1])+abs(c[0]-c[2]) + \
        abs(c[1]-c[0])+abs(c[1]-c[2]) + \
        abs(c[2]-c[1])+abs(c[2]-c[0])


def get_gray_colors(palette):
    list_of_colors = [[hex_to_int(s) for s in color_list_from_hex(c)] for c in palette]
    saturation_list = sorted(
        list_of_colors,
        key=sort_by_saturation
    )
    gray_colors = saturation_list[:(len(saturation_list)//3)]
    gray_colors.sort(key=sum)
    gray_colors = [color_hex_from_list(c) for c in gray_colors]
    return gray_colors


class Plugin(OomoxImportPlugin):

    name = 'import_pil'
    display_name = 'PIL-based images import'
    file_extensions = (
        '.jpg',
        '.png',
        '.gif',
    )

    default_theme = {
        "TERMINAL_THEME_MODE": "manual",
        "TERMINAL_THEME_AUTO_BGFG": False,
    }
    default_themes = {
        "lcars": {
            "TERMINAL_THEME_MODE": "manual",
            "ROUNDNESS": 20,
            "ICONS_STYLE": "archdroid",
        }
    }
    translation_common = {
        "NAME": "scheme",

        "TERMINAL_COLOR0": "color0",
        "TERMINAL_COLOR1": "color1",
        "TERMINAL_COLOR2": "color2",
        "TERMINAL_COLOR3": "color3",
        "TERMINAL_COLOR4": "color4",
        "TERMINAL_COLOR5": "color5",
        "TERMINAL_COLOR6": "color6",
        "TERMINAL_COLOR7": "color7",
        "TERMINAL_COLOR8": "color8",
        "TERMINAL_COLOR9": "color9",
        "TERMINAL_COLOR10": "color10",
        "TERMINAL_COLOR11": "color11",
        "TERMINAL_COLOR12": "color12",
        "TERMINAL_COLOR13": "color13",
        "TERMINAL_COLOR14": "color14",
        "TERMINAL_COLOR15": "color15",

        "TERMINAL_BACKGROUND": "background",
        "TERMINAL_FOREGROUND": "foreground",
        "TERMINAL_ACCENT_COLOR": "color3",
    }
    theme_translations = {
        "1": {
            "BG": "color7",
            "FG": "color0",
            "TXT_BG": "foreground",
            "TXT_FG": "background",
            "SEL_BG": "color4",
            "SEL_FG": "color15",
            "MENU_BG": "background",
            "MENU_FG": "color7",
            "BTN_BG": "color8",
            "BTN_FG": "foreground",
            "HDR_BTN_BG": "background",
            "HDR_BTN_FG": "foreground",

            "ICONS_LIGHT_FOLDER": "color12",
            "ICONS_LIGHT": "color14",
            "ICONS_MEDIUM": "color6",
            "ICONS_DARK": "color4",
        },
        "2": {
            "BG": "color7",
            "FG": "color0",
            "TXT_BG": "foreground",
            "TXT_FG": "color0",
            "SEL_BG": "color1",
            "SEL_FG": "foreground",
            "MENU_BG": "background",
            "MENU_FG": "color7",
            "BTN_BG": "color12",
            "BTN_FG": "color0",
            "WM_BORDER_FOCUS": "color10",
            "ICONS_LIGHT_FOLDER": "color13",
            "ICONS_LIGHT": "color9",
            "ICONS_MEDIUM": "color1",
            "ICONS_DARK": "color5",
        },
        "3": {
            "BG": "color7",
            "FG": "color0",
            "TXT_BG": "foreground",
            "TXT_FG": "color0",
            "SEL_BG": "color3",
            "SEL_FG": "color0",
            "MENU_BG": "background",
            "MENU_FG": "color7",
            "BTN_BG": "color12",
            "BTN_FG": "color0",
            "WM_BORDER_FOCUS": "color3",
            "ICONS_LIGHT_FOLDER": "color13",
            "ICONS_LIGHT": "color9",
            "ICONS_MEDIUM": "color1",
            "ICONS_DARK": "color5",
        },
        "4": {
            "BG": "color5",
            "FG": "color0",
            "MENU_BG": "color4",
            "MENU_FG": "color1",
            "SEL_BG": "color4",
            "SEL_FG": "color0",
            "TXT_BG": "color6",
            "TXT_FG": "color1",
            "BTN_BG": "color3",
            "BTN_FG": "color0",
            "HDR_BTN_BG": "color5",
            "HDR_BTN_FG": "color1",
        },
        "lcars": {
            "BG": "background",
            "FG": "foreground",
            "TXT_BG": "background",
            "TXT_FG": "color11",
            "SEL_BG": "color1",
            "SEL_FG": "background",
            "MENU_BG": "background",
            "MENU_FG": "color7",
            "BTN_BG": "color10",
            "BTN_FG": "color0",
            "HDR_BTN_BG": "color12",
            "HDR_BTN_FG": "background",
            "WM_BORDER_FOCUS": "color10",
            "ICONS_ARCHDROID": "color12",
        },
    }

    theme_model_import = [
        {
            'display_name': _('Import colors from image: '),
            'type': 'separator',
            'value_filter': {
                'FROM_PLUGIN': ['import_pil']
            },
        },
        {
            'key': '_PIL_PALETTE_QUALITY',
            'type': 'options',
            'options': [{
                'value': LOW_QUALITY,
                'display_name': 'low',
            }, {
                'value': MEDIUM_QUALITY,
                'display_name': 'medium',
            }, {
                'value': HIGH_QUALITY,
                'display_name': 'high',
            }, {
                'value': ULTRA_QUALITY,
                'display_name': 'ultra',
            }],
            'fallback_value': LOW_QUALITY,
            'display_name': _('Image analysis quality'),
            'reload_theme': True,
        },
        {
            'key': '_PIL_PALETTE_STYLE',
            'type': 'options',
            'options': [
                {'value': template_name}
                for template_name in sorted(os.listdir(TERMINAL_TEMPLATE_DIR))
            ],
            'fallback_value': 'monovedek_pale_gray',
            'display_name': _('Palette style'),
            'reload_theme': True,
        },
        {
            'key': '_PIL_PALETTE_STRICT',
            'type': 'bool',
            'fallback_value': False,
            'display_name': _('Stronger follow palette template'),
            'reload_theme': True,
        },
        {
            'key': '_PIL_PALETTE_INVERSE',
            'type': 'bool',
            'fallback_value': False,
            'display_name': _('Dark/Light colors'),
            'reload_theme': True,
        },
        {
            'key': '_PIL_THEME_TEMPLATE',
            'type': 'options',
            'options': [
                {'value': template_name}
                for template_name in sorted(theme_translations.keys())
            ],
            'fallback_value': '1',
            'display_name': _('GUI theme template'),
            'reload_theme': True,
        },
        {
            'key': '_PIL_IMAGE_PREVIEW',
            'type': 'image_path',
            'fallback_value': None,
            'display_name': _('Image thumbnail'),
        },
        {
            'display_name': _('Theme options: '),
            'type': 'separator',
            'value_filter': {
                'FROM_PLUGIN': ['import_pil']
            },
        },
    ]

    def read_colorscheme_from_path(self, preset_path):
        from oomox_gui.theme_model import THEME_MODEL_BY_KEY

        THEME_MODEL_BY_KEY['_PIL_IMAGE_PREVIEW']['fallback_value'] = preset_path
        image_palette = self.generate_terminal_palette(
            THEME_MODEL_BY_KEY.get('_PIL_PALETTE_STYLE', {}).get('fallback_value'),
            preset_path
        )
        inverse_palette = bool(
            THEME_MODEL_BY_KEY.get('_PIL_PALETTE_INVERSE', {}).get('fallback_value')
        )

        theme_template = THEME_MODEL_BY_KEY.get('_PIL_THEME_TEMPLATE', {}).get('fallback_value')
        oomox_theme = {}
        oomox_theme.update(self.default_theme)
        if theme_template in self.default_themes:
            oomox_theme.update(self.default_themes[theme_template])
        translation = {}
        translation.update(self.translation_common)
        translation.update(
            self.theme_translations[theme_template]
        )
        for oomox_key, image_palette_key in translation.items():
            if image_palette_key in image_palette:
                oomox_theme[oomox_key] = image_palette[image_palette_key]

        if inverse_palette:
            oomox_theme['TERMINAL_FOREGROUND'], oomox_theme['TERMINAL_BACKGROUND'] = \
                oomox_theme['TERMINAL_BACKGROUND'], oomox_theme['TERMINAL_FOREGROUND']
        return oomox_theme

    _palette_cache = {}
    _terminal_palette_cache = {}

    @classmethod
    def get_image_palette(cls, image_path, quality, use_whole_palette):
        _id = image_path+str(quality)+str(use_whole_palette)
        if not cls._palette_cache.get(_id):
            cls._palette_cache[_id] = image_analyzer.get_hex_palette(
                image_path, quality=quality, use_whole_palette=use_whole_palette
            )
        return cls._palette_cache[_id]

    @classmethod
    def _generate_terminal_palette(cls, template_path, image_path, quality, use_whole_palette):
        hex_palette = cls.get_image_palette(image_path, quality, use_whole_palette)[:]
        gray_colors = get_gray_colors(hex_palette)
        ACCURACY = 40  # pylint: disable=invalid-name
        hex_palette += [hex_darker(c, ACCURACY) for c in gray_colors]
        hex_palette += [hex_darker(c, -ACCURACY) for c in gray_colors]
        reference_palette = import_xcolors(os.path.join(TERMINAL_TEMPLATE_DIR, template_path))
        result_palette = {}
        for key, value in reference_palette.items():
            closest_color, _diff = find_closest_color(value, hex_palette)
            result_palette[key] = closest_color
        gc.collect()
        return result_palette

    @classmethod
    def generate_terminal_palette(cls, template_path, image_path):
        from oomox_gui.theme_model import THEME_MODEL_BY_KEY
        quality = int(
            THEME_MODEL_BY_KEY.get('_PIL_PALETTE_QUALITY', {}).get('fallback_value')
        )
        use_whole_palette = bool(
            THEME_MODEL_BY_KEY.get('_PIL_PALETTE_STRICT', {}).get('fallback_value')
        )
        _id = template_path+image_path+str(quality)+str(use_whole_palette)
        if not cls._terminal_palette_cache.get(_id):
            cls._terminal_palette_cache[_id] = cls._generate_terminal_palette(
                template_path, image_path, quality, use_whole_palette
            )
        palette = {}
        palette.update(cls._terminal_palette_cache[_id])
        return palette



import os
from enum import Enum

from gi.repository import Gtk

from .gtk_helpers import ScaledImage


class IconsNames(Enum):
    HOME = 'user-home'
    DESKTOP = 'user-desktop'
    FILE_MANAGER = 'system-file-manager'


class IconThemePreview(Gtk.ListBox):

    icons_plugin_name = None

    icons_templates = None
    icons_imageboxes = None

    def __init__(self):
        self.icons_imageboxes = {}
        self.icons_templates = {}

        super().__init__()
        self.set_margin_left(10)
        self.set_margin_right(10)
        self.set_selection_mode(Gtk.SelectionMode.NONE)

        row = Gtk.ListBoxRow()
        hbox = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=6)
        row.add(hbox)
        for icon in IconsNames:
            icon_imagebox = ScaledImage(width=48)
            hbox.pack_start(icon_imagebox, True, True, 0)
            self.icons_imageboxes[icon.name] = icon_imagebox
        self.add(row)
        self.show_all()

    def update_preview(self, colorscheme, theme_plugin):
        transform_function = theme_plugin.preview_transform_function
        self.load_icon_templates(theme_plugin)
        for icon in IconsNames:
            new_svg_image = transform_function(
                self.icons_templates[icon.name],
                colorscheme
            ).encode('ascii')
            self.icons_imageboxes[icon.name].set_from_bytes(new_svg_image)

    def load_icon_templates(self, theme_plugin):
        if theme_plugin.name == self.icons_plugin_name:
            return
        self.icons_plugin_name = theme_plugin.name
        for icon in IconsNames:
            template_path = "{}.svg.template".format(icon.value)
            with open(
                    os.path.join(
                        theme_plugin.preview_svg_dir, template_path
                    ), "rb"
            ) as file_object:
                self.icons_templates[icon.name] = file_object.read().decode('utf-8')



import os

from .config import PLUGINS_DIR, USER_PLUGINS_DIR
from .plugin_api import (
    OomoxPlugin,
    OomoxImportPlugin, OomoxThemePlugin, OomoxIconsPlugin, OomoxExportPlugin,
)
from .helpers import get_plugin_module


ALL_PLUGINS = {}
THEME_PLUGINS = {}
ICONS_PLUGINS = {}
EXPORT_PLUGINS = {}
IMPORT_PLUGINS = {}


def init_plugins():
    all_plugin_paths = {}
    for _plugins_dir in (PLUGINS_DIR, USER_PLUGINS_DIR):
        if not os.path.exists(_plugins_dir):
            continue
        for plugin_name in os.listdir(_plugins_dir):
            all_plugin_paths[plugin_name] = os.path.join(_plugins_dir, plugin_name)

    for plugin_name, plugin_path in all_plugin_paths.items():
        plugin_module = get_plugin_module(
            plugin_name,
            os.path.join(plugin_path, "oomox_plugin.py")
        )
        plugin_class = plugin_module.Plugin
        plugin = plugin_class()
        if not issubclass(plugin_class, OomoxPlugin):
            continue
        ALL_PLUGINS[plugin_name] = plugin
        if issubclass(plugin_class, OomoxImportPlugin):
            IMPORT_PLUGINS[plugin_name] = plugin
        if issubclass(plugin_class, OomoxThemePlugin):
            THEME_PLUGINS[plugin_name] = plugin
        if issubclass(plugin_class, OomoxIconsPlugin):
            ICONS_PLUGINS[plugin_name] = plugin
        if issubclass(plugin_class, OomoxExportPlugin):
            EXPORT_PLUGINS[plugin_name] = plugin


init_plugins()


def _print_debug_plugins():
    # @TODO: remove debug code:
    print("MAIN:")
    print()
    print("import plugins:")
    print(IMPORT_PLUGINS)
    print("theme plugins:")
    print(THEME_PLUGINS)
    print("icons plugins:")
    print(ICONS_PLUGINS)
    print("export plugins:")
    print(EXPORT_PLUGINS)
    print()
    print("all plugins:")
    print(ALL_PLUGINS)
    print()
    for plugin_name, plugin in ALL_PLUGINS.items():
        print(
            "{}: {}".format(plugin_name, plugin.display_name)
        )


if __name__ == "__main__":
    _print_debug_plugins()



import os

from gi.repository import Gtk

from oomox_gui.export_common import OPTION_GTK2_HIDPI, GtkThemeExportDialog
from oomox_gui.plugin_api import OomoxThemePlugin
from oomox_gui.i18n import _


OPTION_GTK3_CURRENT_VERSION_ONLY = 'OPTION_GTK3_CURRENT_VERSION_ONLY'
OPTION_EXPORT_CINNAMON_THEME = 'OPTION_EXPORT_CINNAMON_THEME'


PLUGIN_DIR = os.path.dirname(os.path.realpath(__file__))
GTK_THEME_DIR = os.path.join(PLUGIN_DIR, "gtk-theme/")


class OomoxThemeExportDialog(GtkThemeExportDialog):
    timeout = 100
    config_name = 'gtk_theme_oomox'

    def do_export(self):
        self.command = [
            "bash",
            os.path.join(GTK_THEME_DIR, "change_color.sh"),
            "--hidpi", str(self.export_config[OPTION_GTK2_HIDPI]),
            "--output", self.theme_name,
            self.temp_theme_path,
        ]
        make_opts = []
        if self.export_config[OPTION_GTK3_CURRENT_VERSION_ONLY]:
            if Gtk.get_minor_version() >= 20:
                make_opts += ["gtk320"]
            else:
                make_opts += ["gtk3"]
        else:
            make_opts += ["gtk3", "gtk320"]
        if self.export_config[OPTION_EXPORT_CINNAMON_THEME]:
            make_opts += ["css_cinnamon"]
        if make_opts:
            self.command += [
                "--make-opts", " ".join(make_opts),
            ]
        super().do_export()

    def __init__(self, transient_for, colorscheme, theme_name, **kwargs):
        super().__init__(
            transient_for=transient_for,
            colorscheme=colorscheme,
            theme_name=theme_name,
            add_options={
                OPTION_GTK3_CURRENT_VERSION_ONLY: {
                    'default': False,
                    'display_name': _("Generate theme only for the current _GTK+3 version\n"
                                      "instead of both 3.18 and 3.20+"),
                },
                OPTION_EXPORT_CINNAMON_THEME: {
                    'default': False,
                    'display_name': _("Generate theme for _Cinnamon"),
                },
            },
            **kwargs
        )


class Plugin(OomoxThemePlugin):

    name = 'oomox'
    display_name = 'Numix-based'
    description = '(GTK+2, GTK+3, Cinnamon, Metacity, Openbox, Qt5ct, Unity, Xfwm)'
    export_dialog = OomoxThemeExportDialog
    gtk_preview_dir = os.path.join(PLUGIN_DIR, "gtk_preview_css/")

    enabled_keys_gtk = [
        'BG',
        'FG',
        'MENU_BG',
        'MENU_FG',
        'SEL_BG',
        'SEL_FG',
        'ACCENT_BG',
        'TXT_BG',
        'TXT_FG',
        'BTN_BG',
        'BTN_FG',
        'HDR_BTN_BG',
        'HDR_BTN_FG',
        'WM_BORDER_FOCUS',
        'WM_BORDER_UNFOCUS',
    ]

    enabled_keys_options = [
        'ROUNDNESS',
        'SPACING',
        'GRADIENT',
        'GTK3_GENERATE_DARK',
    ]

    theme_model_options = [
        {
            'key': 'OUTLINE_WIDTH',
            'type': 'int',
            'fallback_value': 1,
            'display_name': _('(GTK3) Focused outline width'),
        },
        {
            'key': 'BTN_OUTLINE_WIDTH',
            'type': 'int',
            'fallback_value': 1,
            'display_name': _('(GTK3) Focused button outline width'),
        },
        {
            'key': 'BTN_OUTLINE_OFFSET',
            'type': 'int',
            'fallback_value': -3,
            'min_value': -20,
            'display_name': _('(GTK3) Focused button outline offset'),
        },
        # {  @TODO: implement value_filter for separators
        #     'type': 'separator',
        #     'display_name': _('Text input caret'),
        # },
        {
            'key': 'CARET1_FG',
            'type': 'color',
            'fallback_key': 'TXT_FG',
            'display_name': _('Primary caret color'),
        },
        {
            'key': 'CARET2_FG',
            'type': 'color',
            'fallback_key': 'TXT_FG',
            'display_name': _('Secondary caret color'),
        },
        {
            'key': 'CARET_SIZE',
            'type': 'float',
            'fallback_value': 0.04,  # GTK's default
            'display_name': _('Caret aspect ratio'),
        },
    ]

    theme_model_other = [
        {
            'key': 'UNITY_DEFAULT_LAUNCHER_STYLE',
            'type': 'bool',
            'fallback_value': False,
            'display_name': _('(Unity) Use default launcher style'),
        },
        {
            'key': 'CINNAMON_OPACITY',
            'type': 'float',
            'fallback_value': 1.0,
            'max_value': 1.0,
            'display_name': _('(Cinnamon) Opacity'),
        },
    ]

    theme_model_extra = [
    ]

    def preview_before_load_callback(self, preview_object, colorscheme):
        preview_object.WM_BORDER_WIDTH = 2



from enum import Enum

from gi.repository import Gtk, Gio, GLib, GdkPixbuf
from gi.types import GObjectMeta

from .i18n import _


class ActionsEnum(Enum):

    def get_name(self):
        return self.name

    def get_id(self):
        return '.'.join([self._target.value, self.name])


class CenterLabel(Gtk.Label):
    def __init__(self, label=None):
        super().__init__()
        if label:
            self.set_text(label)
        self.set_justify(Gtk.Justification.CENTER)
        self.set_alignment(0.5, 0.5)
        self.set_margin_left(6)
        self.set_margin_right(6)
        self.set_margin_top(6)
        self.set_margin_bottom(6)


class ImageContainer(Gtk.Container):
    icon = None
    image = None

    def __init__(self, icon_name, tooltip_text=None):
        super().__init__()
        self.icon = Gio.ThemedIcon(name=icon_name)
        self.image = Gtk.Image.new_from_gicon(self.icon, Gtk.IconSize.BUTTON)
        self.add(self.image)
        if tooltip_text:
            self.set_tooltip_text(tooltip_text)


class ImageButton(Gtk.Button, ImageContainer):
    def __init__(self, *args, **kwargs):
        Gtk.Button.__init__(self)
        ImageContainer.__init__(self, *args, **kwargs)


class ImageMenuButton(Gtk.MenuButton, ImageContainer):
    def __init__(self, *args, **kwargs):
        Gtk.MenuButton.__init__(self)
        ImageContainer.__init__(self, *args, **kwargs)


class ScaledImage(Gtk.Image):

    scale_factor = None
    orig_width = None
    orig_height = None
    oomox_width = None
    oomox_height = None

    def __init__(self, *args, width=None, height=None, **kwargs):
        super().__init__(*args, **kwargs)
        if not width or height:
            raise TypeError('Either "width" or "height" should be set')
        self._set_orig_dimensions(width=width, height=height)
        style_context = self.get_style_context()
        self.scale_factor = style_context.get_scale()

    def _set_orig_dimensions(self, width=None, height=None):
        if width:
            self.orig_width = width
        if height:
            self.orig_height = height

    def do_draw(self, cr):  # pylint: disable=arguments-differ
        if self.oomox_width:
            cr.scale(1/self.scale_factor, 1/self.scale_factor)
            cr.translate(
                self.oomox_width - self.oomox_width/self.scale_factor,
                self.oomox_height - self.oomox_height/self.scale_factor
            )
            Gtk.Image.do_draw(self, cr)

    def do_get_preferred_width(self):  # pylint: disable=arguments-differ
        if self.oomox_width:
            return self.oomox_width, self.oomox_width
        return Gtk.Image.do_get_preferred_width(self)

    def do_get_preferred_height(self):  # pylint: disable=arguments-differ
        if self.oomox_height:
            return self.oomox_height, self.oomox_height
        return Gtk.Image.do_get_preferred_height(self)

    def set_from_bytes(self, bytes_sequence, width=None, height=None):
        self._set_orig_dimensions(width=width, height=height)
        stream = Gio.MemoryInputStream.new_from_bytes(
            GLib.Bytes.new(bytes_sequence)
        )
        self.oomox_width = self.orig_width
        self.oomox_height = self.orig_height

        # @TODO: is it possible to make it faster?
        pixbuf = GdkPixbuf.Pixbuf.new_from_stream_at_scale(
            stream,
            self.oomox_width*self.scale_factor if self.oomox_width else -1,
            self.oomox_height*self.scale_factor if self.oomox_height else -1,
            True,
            None
        )
        self.oomox_width = pixbuf.props.width // self.scale_factor
        self.oomox_height = pixbuf.props.height // self.scale_factor
        self.set_from_pixbuf(pixbuf)


class EntryDialog(Gtk.Dialog):

    entry = None
    entry_text = ''

    def do_response(self, response):  # pylint: disable=arguments-differ
        if response == Gtk.ResponseType.OK:
            self.entry_text = self.entry.get_text()
        self.destroy()

    def __init__(
            self, transient_for,
            title, text, entry_text=None
    ):
        super().__init__(
            title=title,
            transient_for=transient_for,
            flags=0
        )

        self.set_default_size(150, 100)

        label = Gtk.Label(label=text)
        self.entry = Gtk.Entry()
        self.entry.set_activates_default(True)
        if entry_text:
            self.entry.set_text(entry_text)

        box = self.get_content_area()
        box.add(label)
        box.add(self.entry)

        self.add_button(_("_Cancel"), Gtk.ResponseType.CANCEL)
        self.add_button(_("_OK"), Gtk.ResponseType.OK)

        self.set_default_response(Gtk.ResponseType.OK)

        self.show_all()


class YesNoDialog(Gtk.Dialog):

    def do_response(self, _response):  # pylint: disable=arguments-differ
        self.destroy()

    def __init__(self, transient_for,
                 title="",
                 text=_("Are you sure?"),
                 default_response=Gtk.ResponseType.NO):
        super().__init__(
            title=title,
            transient_for=transient_for,
            flags=0
        )
        self.set_default_size(150, 100)

        label = CenterLabel(label=text)
        box = self.get_content_area()
        box.add(label)

        self.add_button(_("_No"), Gtk.ResponseType.NO)
        self.add_button(_("_Yes"), Gtk.ResponseType.YES)

        self.set_default_response(default_response)

        self.show_all()


class GObjectABCMetaAbstractProperty():
    pass


class GObjectABCMeta(GObjectMeta):

    ABS_METHODS = '__abstract_methods__'

    def __init__(cls, name, transient_fors, data):
        super().__init__(name, transient_fors, data)
        for property_name in dir(cls):
            if getattr(cls, property_name) is GObjectABCMetaAbstractProperty:
                setattr(
                    cls, cls.ABS_METHODS,
                    getattr(cls, cls.ABS_METHODS, []) + [property_name]
                )
                delattr(cls, property_name)
        if not (getattr(cls, cls.ABS_METHODS) and not any(
                cls.ABS_METHODS in B.__dict__ for B in cls.__mro__[1:]
        )):
            required_methods = getattr(cls, cls.ABS_METHODS)
            for method_name in required_methods:
                if any(method_name in B.__dict__ for B in cls.__mro__):
                    return
            raise TypeError(
                "Can't instantiate abstract class {} with abstract methods {}".format(
                    cls.__name__,
                    ','.join(required_methods)
                )
            )


def g_abstractproperty(_function):
    return GObjectABCMetaAbstractProperty



from gi.repository import Gtk

from .color import (
    mix_gdk_colors, convert_gdk_to_theme_color
)
from .theme_file import (
    is_user_colorscheme, get_presets
)
from .i18n import _


class ThemePresetsList(Gtk.Box):

    treeview_default_fg = None
    update_signal = None
    treestore = None
    treeview = None
    preset_select_callback = None

    DISPLAY_NAME = 0
    THEME_NAME = 1
    THEME_PATH = 2

    def on_preset_select(self, widget):
        treepath = widget.get_cursor()[0]
        if not treepath:
            return
        treeiter = self.treestore.get_iter(treepath)
        current_theme = self.treestore.get_value(treeiter, self.THEME_NAME)
        current_preset_path = self.treestore.get_value(treeiter, self.THEME_PATH)
        self.preset_select_callback(
            current_theme, current_preset_path
        )

    def add_preset(self, preset_name, preset_path, display_name=None):
        if not display_name:
            display_name = preset_name
        is_default = not is_user_colorscheme(preset_path)
        self.treestore.append(None, (
            display_name, preset_name, preset_path,
            self.treeview_default_fg if is_default else None
        ))

    def _find_treepath_by_filepath(
            self, store, target_filepath, treeiter=None
    ):
        if not treeiter:
            treeiter = self.treestore.get_iter_first()
        while treeiter:
            current_filepath = store[treeiter][self.THEME_PATH]
            if current_filepath == target_filepath:
                return store[treeiter].path
            if store.iter_has_child(treeiter):
                childiter = store.iter_children(treeiter)
                child_result = self._find_treepath_by_filepath(
                    store, target_filepath, childiter
                )
                if child_result:
                    return child_result
            treeiter = store.iter_next(treeiter)
        return None

    def focus_preset_by_filepath(self, filepath):
        treepath = self._find_treepath_by_filepath(
            self.treestore, filepath
        )
        if treepath:
            treepathcopy = treepath.copy()
            while treepath.up():
                self.treeview.expand_row(treepath, False)
            self.treeview.set_cursor(treepathcopy)

    def load_presets(self):
        if self.update_signal:
            self.treeview.disconnect(self.update_signal)
        self.treestore.clear()
        for preset_dir, preset_list in get_presets().items():
            sorted_preset_list = sorted(preset_list, key=lambda x: x['name'])
            first_preset = sorted_preset_list[0]
            color = self.treeview_default_fg if first_preset['default'] \
                else None
            piter = self.treestore.append(
                None, (
                    preset_dir,
                    first_preset['name'],
                    first_preset['path'],
                    color
                )
            )
            for preset in sorted_preset_list[1:]:
                self.treestore.append(
                    piter,
                    (preset['name'], preset['name'], preset['path'], color)
                )
        self.update_signal = self.treeview.connect(
            "cursor_changed", self.on_preset_select
        )

    def _get_color_for_default_preset_names(self):
        treeview_style_context = self.treeview.get_style_context()
        treeview_fg = treeview_style_context.get_color(
            Gtk.StateFlags.NORMAL
        )
        treeview_bg = treeview_style_context.get_background_color(
            Gtk.StateFlags.NORMAL
        )
        self.treeview_default_fg = '#' + convert_gdk_to_theme_color(
            mix_gdk_colors(treeview_fg, treeview_bg, 0.5)
        )

    def __init__(self, preset_select_callback):
        super().__init__(orientation=Gtk.Orientation.VERTICAL)
        self.preset_select_callback = preset_select_callback

        self.treestore = Gtk.TreeStore(str, str, str, str)
        self.treestore.set_sort_column_id(0, Gtk.SortType.ASCENDING)
        self.treeview = Gtk.TreeView(
            model=self.treestore, headers_visible=False
        )
        self._get_color_for_default_preset_names()
        column = Gtk.TreeViewColumn(
            cell_renderer=Gtk.CellRendererText(), text=0, foreground=3
        )
        self.treeview.append_column(column)
        self.load_presets()

        scrolled = Gtk.ScrolledWindow()
        scrolled.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.AUTOMATIC)
        scrolled.add(self.treeview)

        presets_list_label = Gtk.Label()
        presets_list_label.set_text(_("Presets:"))
        self.pack_start(presets_list_label, False, False, 0)
        self.pack_start(scrolled, True, True, 0)



import subprocess


class XrdbCache():

    _cache = None

    @classmethod
    def get(cls):
        if cls._cache:
            return cls._cache

        timeout = 10
        command = ['xrdb', '-query']

        result = {}
        proc = subprocess.Popen(
            command,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT
        )
        for line in iter(proc.stdout.readline, b''):
            line = line.decode("utf-8")
            key, value, *_rest = line.split(':')
            key = key.lstrip('*').lstrip('.')
            value = value.strip()
            result[key] = value
        proc.communicate(timeout=timeout)
        if proc.returncode == 0:
            cls._cache = result
            return result
        print('xrdb not found')
        return None

    @classmethod
    def clear(cls):
        cls._cache = None



#!/bin/env python3
import sys
import signal

from gi.repository import Gtk, Gio

from .i18n import _
from .config import USER_COLORS_DIR
from .helpers import mkdir_p
from .gtk_helpers import (
    ImageButton, ImageMenuButton,
    EntryDialog, YesNoDialog,
    ActionsEnum,
)
from .theme_file import (
    get_user_theme_path, is_user_colorscheme, is_colorscheme_exists,
    save_colorscheme, remove_colorscheme,
)
from .theme_file_parse import read_colorscheme_from_path
from .presets_list import ThemePresetsList
from .colors_list import ThemeColorsList
from .preview import ThemePreview
from .export_common import export_terminal_theme
from .terminal import generate_terminal_colors_for_oomox
from .plugin_loader import THEME_PLUGINS, ICONS_PLUGINS, EXPORT_PLUGINS


class NewDialog(EntryDialog):

    def __init__(
            self, transient_for,
            title=_("New theme"),
            text=_("Please input new theme name:"),
            entry_text=None
    ):
        super().__init__(
            transient_for=transient_for,
            title=title,
            text=text,
            entry_text=entry_text
        )


class RenameDialog(NewDialog):

    def __init__(self, transient_for, entry_text):
        super().__init__(
            transient_for=transient_for,
            title=_("Rename theme"),
            entry_text=entry_text
        )


class UnsavedDialog(YesNoDialog):

    def __init__(self, transient_for):
        super().__init__(
            transient_for=transient_for,
            title=_("Unsaved changes"),
            text=_("There are unsaved changes.\nSave them?")
        )


class RemoveDialog(YesNoDialog):

    def __init__(self, transient_for):
        super().__init__(
            transient_for=transient_for,
            title=_("Remove theme"),
            text=_("Are you sure you want to delete the colorscheme?\n"
                   "This can not be undone.")
        )


def dialog_is_yes(dialog):
    return dialog.run() in (Gtk.ResponseType.YES, Gtk.ResponseType.OK)


class AppActions(ActionsEnum):
    _target = "app"
    quit = "quit"


class WindowActions(ActionsEnum):
    _target = "win"
    clone = "clone"
    export_icons = "icons"
    export_theme = "theme"
    export_terminal = "terminal"
    menu = "menu"
    remove = "remove"
    rename = "rename"
    save = "save"


class WindowWithActions(Gtk.ApplicationWindow):

    def action_tooltip(self, action_enum, tooltip):
        action_id = action_enum.get_id()
        accels = self.get_application().get_accels_for_action(action_id)
        if accels:
            key, mods = Gtk.accelerator_parse(accels[0])
            tooltip += ' ({})'.format(Gtk.accelerator_get_label(key, mods))
        return tooltip

    def attach_action(self, widget, action_enum, with_tooltip=True):
        action_id = action_enum.get_id()
        widget.set_action_name(action_id)
        if with_tooltip:
            tooltip = self.action_tooltip(action_enum, widget.get_tooltip_text())
            widget.set_tooltip_text(tooltip)

    def add_simple_action_by_name(self, action_name, callback):
        action = Gio.SimpleAction.new(action_name, None)
        action.connect("activate", callback)
        self.add_action(action)
        return action

    def add_simple_action(self, action_enum, callback):
        return self.add_simple_action_by_name(action_enum.name, callback)


class OomoxApplicationWindow(WindowWithActions):  # pylint: disable=too-many-instance-attributes

    colorscheme_name = None
    colorscheme_path = None
    colorscheme = None
    colorscheme_is_user = None
    theme_edited = False
    #
    plugin_theme = None
    plugin_icons = None
    #
    # actions:
    save_action = None
    rename_action = None
    remove_action = None
    #
    # widget sections:
    box = None
    headerbar = None
    theme_edit = None
    presets_list = None
    preview = None

    def save_theme(self, name=None):
        if not name:
            name = self.colorscheme_name
        if not is_user_colorscheme(self.colorscheme_path):
            if self.check_colorscheme_exists(name):
                self.clone_theme()
                return
        new_path = save_colorscheme(name, self.colorscheme)
        self.theme_edited = False
        old_path = self.colorscheme_path
        self.colorscheme_name = name
        self.colorscheme_path = new_path
        if old_path != new_path:
            self.reload_presets()
        self.save_action.set_enabled(False)
        self.headerbar.props.title = self.colorscheme_name

    def remove_theme(self, name=None):
        if not name:
            name = self.colorscheme_name
        try:
            remove_colorscheme(name)
        except FileNotFoundError:
            pass

    def clone_theme(self):
        dialog = NewDialog(transient_for=self)
        if not dialog_is_yes(dialog):
            return
        new_theme_name = dialog.entry_text
        if not self.check_colorscheme_exists(new_theme_name):
            self.save_theme(new_theme_name)
            self.reload_presets()
        else:
            self.clone_theme()

    def rename_theme(self):
        dialog = RenameDialog(transient_for=self, entry_text=self.colorscheme_name)
        if not dialog_is_yes(dialog):
            return
        new_theme_name = dialog.entry_text
        if not self.check_colorscheme_exists(new_theme_name):
            old_theme_name = self.colorscheme_name
            self.save_theme(new_theme_name)
            self.remove_theme(old_theme_name)
            self.reload_presets()

    def check_unsaved_changes(self):
        if self.theme_edited:
            if dialog_is_yes(UnsavedDialog(transient_for=self)):
                self.save_theme()

    def check_colorscheme_exists(self, colorscheme_name):
        if not is_colorscheme_exists(get_user_theme_path(colorscheme_name)):
            return False
        dialog = Gtk.MessageDialog(
            transient_for=self,
            flags=0,
            message_type=Gtk.MessageType.WARNING,
            buttons=Gtk.ButtonsType.OK,
            message_format=_("Colorscheme with such name already exists")
        )
        dialog.run()
        dialog.destroy()
        return True

    def select_theme_plugin(self):
        theme_plugin_name = self.colorscheme['THEME_STYLE']
        self.plugin_theme = None
        for theme_plugin in THEME_PLUGINS.values():
            if theme_plugin.name == theme_plugin_name:
                self.plugin_theme = theme_plugin

    def select_icons_plugin(self):
        icons_plugin_name = self.colorscheme['ICONS_STYLE']
        self.plugin_icons = None
        for icons_plugin in ICONS_PLUGINS.values():
            if icons_plugin.name == icons_plugin_name:
                self.plugin_icons = icons_plugin

    def load_colorscheme(self, colorscheme):
        self.colorscheme = colorscheme
        self.select_theme_plugin()
        self.select_icons_plugin()
        self.generate_terminal_colors()
        try:
            self.preview.update_preview(
                colorscheme=self.colorscheme,
                theme_plugin=self.plugin_theme,
                icons_plugin=self.plugin_icons,
            )
        except Exception as exc:
            import traceback
            print()
            print("ERROR: Can't show theme preview:")
            print(exc)
            traceback.print_exc()
            print()

    def on_preset_selected(self, selected_preset, selected_preset_path):
        self.check_unsaved_changes()
        self.colorscheme_name = selected_preset
        self.colorscheme_path = selected_preset_path
        self.load_colorscheme(read_colorscheme_from_path(selected_preset_path))
        self.colorscheme_is_user = is_user_colorscheme(self.colorscheme_path)
        self.theme_edit.open_theme(self.colorscheme)
        self.theme_edited = False
        self.save_action.set_enabled(False)
        self.rename_action.set_enabled(self.colorscheme_is_user)
        self.remove_action.set_enabled(self.colorscheme_is_user)
        self.headerbar.props.title = selected_preset

    def theme_reload(self):
        self.on_preset_selected(
            self.colorscheme_name, self.colorscheme_path
        )
        return self.colorscheme

    def generate_terminal_colors(self):
        self.colorscheme.update(generate_terminal_colors_for_oomox(self.colorscheme))

    def on_color_edited(self, colorscheme):
        self.generate_terminal_colors()
        self.load_colorscheme(colorscheme)
        if not self.theme_edited:
            self.headerbar.props.title = "*" + self.colorscheme_name
            self.save_action.set_enabled(True)
        self.theme_edited = True

    def reload_presets(self):
        self.presets_list.load_presets()
        self.presets_list.focus_preset_by_filepath(self.colorscheme_path)

    ###########################################################################
    # Signal handlers:
    ###########################################################################

    def _on_clone(self, _action, _param=None):
        return self.clone_theme()

    def _on_rename(self, _action, _param=None):
        self.rename_theme()

    def _on_remove(self, _action, _param=None):
        if not dialog_is_yes(RemoveDialog(transient_for=self)):
            return
        self.remove_theme()
        self.reload_presets()

    def _on_save(self, _action, _param=None):
        self.save_theme()

    def _on_export_theme(self, _action, _param=None):
        self.plugin_theme.export_dialog(
            transient_for=self,
            theme_name=self.colorscheme_name,
            colorscheme=self.colorscheme
        )

    def _on_export_icontheme(self, _action, _param=None):
        self.plugin_icons.export_dialog(
            transient_for=self,
            theme_name=self.colorscheme_name,
            colorscheme=self.colorscheme
        )

    def _on_export_terminal(self, _action, _param=None):
        export_terminal_theme(transient_for=self, colorscheme=self.colorscheme)

    def _on_export_plugin(self, action, _param=None):
        plugin = EXPORT_PLUGINS[
            action.props.name.replace('export_plugin_', '')
        ]
        plugin.export_dialog(
            transient_for=self,
            theme_name=self.colorscheme_name,
            colorscheme=self.colorscheme
        )

    def _on_quit(self, _arg1, _arg2):
        self.check_unsaved_changes()

    ###########################################################################
    # Init widgets:
    ###########################################################################

    def _init_headerbar(self):
        self.headerbar = Gtk.HeaderBar()
        self.headerbar.set_show_close_button(True)
        self.headerbar.props.title = _("Oo-mox GUI")

        # @TODO:
        # new_button = ImageButton("text-x-generic-symbolic", _("Create new theme"))  # noqa
        # self.headerbar.pack_start(new_button)

        clone_button = ImageButton(
            "edit-copy-symbolic", _("Clone current theme")
        )
        self.attach_action(clone_button, WindowActions.clone)
        self.headerbar.pack_start(clone_button)

        save_button = ImageButton(
            "document-save-symbolic", _("Save theme")
        )
        self.attach_action(save_button, WindowActions.save)
        self.headerbar.pack_start(save_button)

        rename_button = ImageButton(
            # "preferences-desktop-font-symbolic", "Rename theme"
            "pda-symbolic", _("Rename theme")
        )
        self.attach_action(rename_button, WindowActions.rename)
        self.headerbar.pack_start(rename_button)

        remove_button = ImageButton(
            "edit-delete-symbolic", _("Remove theme")
        )
        self.attach_action(remove_button, WindowActions.remove)
        self.headerbar.pack_start(remove_button)

        #

        if EXPORT_PLUGINS:
            menu = Gio.Menu()
            for plugin_name, plugin in EXPORT_PLUGINS.items():
                menu.append_item(Gio.MenuItem.new(
                    plugin.display_name,
                    "win.export_plugin_{}".format(plugin_name)
                ))

            menu_button = ImageMenuButton("open-menu-symbolic")
            menu_button.set_use_popover(True)
            menu_button.set_menu_model(menu)
            self.add_action(Gio.PropertyAction(
                name=WindowActions.get_name(WindowActions.menu),
                object=menu_button,
                property_name="active"
            ))
            self.headerbar.pack_end(menu_button)

        export_terminal_button = Gtk.Button(
            label=_("Export _terminal"),
            use_underline=True,
            tooltip_text=_("Export terminal theme")
        )
        self.attach_action(export_terminal_button, WindowActions.export_terminal)
        self.headerbar.pack_end(export_terminal_button)

        export_icons_button = Gtk.Button(label=_("Export _icons"),
                                         use_underline=True,
                                         tooltip_text=_("Export icon theme"))
        self.attach_action(export_icons_button, WindowActions.export_icons)
        self.headerbar.pack_end(export_icons_button)

        export_button = Gtk.Button(label=_("_Export theme"),
                                   use_underline=True,
                                   tooltip_text=_("Export GTK theme"))
        self.attach_action(export_button, WindowActions.export_theme)
        self.headerbar.pack_end(export_button)

        self.set_titlebar(self.headerbar)

    def _init_window(self):
        self.set_wmclass("oomox", "Oomox")
        self.connect("delete-event", self._on_quit)
        self.set_default_size(500, 300)
        self.set_border_width(6)

        self._init_headerbar()

        self.box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=6)
        self.add(self.box)

    def _init_actions(self):
        self.add_simple_action(WindowActions.clone, self._on_clone)
        self.save_action = self.add_simple_action(WindowActions.save, self._on_save)
        self.rename_action = self.add_simple_action(WindowActions.rename, self._on_rename)
        self.remove_action = self.add_simple_action(WindowActions.remove, self._on_remove)
        self.add_simple_action(WindowActions.export_theme, self._on_export_theme)
        self.add_simple_action(WindowActions.export_icons, self._on_export_icontheme)
        self.add_simple_action(WindowActions.export_terminal, self._on_export_terminal)
        for plugin_name in EXPORT_PLUGINS:
            self.add_simple_action_by_name(
                "export_plugin_{}".format(plugin_name), self._on_export_plugin
            )

    def __init__(self, application, title=_("Oo-mox GUI")):
        super().__init__(
            application=application, title=title
        )
        self.colorscheme = {}
        mkdir_p(USER_COLORS_DIR)

        self._init_actions()
        self._init_window()

        self.presets_list = ThemePresetsList(
            preset_select_callback=self.on_preset_selected
        )
        self.box.pack_start(self.presets_list, False, False, 0)

        self.theme_edit = ThemeColorsList(
            color_edited_callback=self.on_color_edited,
            theme_reload_callback=self.theme_reload,
            transient_for=self
        )
        self.box.pack_start(self.theme_edit, True, True, 0)

        self.preview = ThemePreview()
        self.box.pack_start(self.preview, False, False, 0)

        self.show_all()


class OomoxGtkApplication(Gtk.Application):

    window = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, application_id="com.github.themix_project.Oomox",
                         flags=Gio.ApplicationFlags.HANDLES_COMMAND_LINE,
                         **kwargs)
        # @TODO: use oomox-gui as the only one entrypoint to all cli tools
        # self.add_main_option("test", ord("t"), GLib.OptionFlags.NONE,
        # GLib.OptionArg.NONE, "Command line test", None)

    def do_startup(self):  # pylint: disable=arguments-differ

        def set_accels_for_action(action, accels):
            self.set_accels_for_action(action.get_id(), accels)

        Gtk.Application.do_startup(self)

        quit_action = Gio.SimpleAction.new(
            AppActions.get_name(AppActions.quit), None
        )
        quit_action.connect("activate", self._on_quit)
        self.add_action(quit_action)

        set_accels_for_action(AppActions.quit, ["<Primary>Q"])
        set_accels_for_action(WindowActions.clone, ["<Primary>D"])
        set_accels_for_action(WindowActions.save, ["<Primary>S"])
        set_accels_for_action(WindowActions.rename, ["F2"])
        set_accels_for_action(WindowActions.remove, ["<Primary>Delete"])
        set_accels_for_action(WindowActions.export_theme, ["<Primary>E"])
        set_accels_for_action(WindowActions.export_icons, ["<Primary>I"])
        set_accels_for_action(WindowActions.menu, ["F10"])

    def do_activate(self):  # pylint: disable=arguments-differ
        if not self.window:
            self.window = OomoxApplicationWindow(application=self)
        self.window.present()

    def do_command_line(self, _command_line):  # pylint: disable=arguments-differ
        # options = command_line.get_options_dict()
        # if options.contains("test"):
            # print("Test argument recieved")
        self.activate()
        return 0

    def quit(self):  # pylint: disable=arguments-differ
        if self.window:
            self.window.close()
        else:
            super().quit()

    def _on_quit(self, _action, _param=None):
        self.quit()


def main():

    app = OomoxGtkApplication()

    def handle_sig_int(*_whatever):  # pragma: no cover
        app.quit()
        sys.stderr.write("\n\nCanceled by user (SIGINT)\n")
        sys.exit(125)

    signal.signal(signal.SIGINT, handle_sig_int)
    app.run(sys.argv)


if __name__ == "__main__":
    main()



import subprocess
import os
import tempfile
from threading import Thread

from gi.repository import Gtk, GLib, Pango

from .export_config import ExportConfig
from .theme_file import save_colorscheme
from .terminal import (
    generate_terminal_colors_for_oomox,
    generate_xrdb_theme_from_oomox,
    generate_xresources
)
from .gtk_helpers import CenterLabel
from .i18n import _


class ExportDialog(Gtk.Dialog):

    command = None
    timeout = 120

    # widgets:
    box = None
    label = None
    spinner = None
    options_box = None
    scrolled_window = None
    log = None
    error_box = None
    apply_button = None

    def _close_button_callback(self, _widget):
        self.destroy()

    def show_error(self):
        self.box.remove(self.label)
        self.box.remove(self.spinner)

        error_label = CenterLabel(
            label=_("Something went wrong :(")
        )
        error_label.set_alignment(0.5, 0.5)

        error_dismiss_button = Gtk.Button(label=_("_Dismiss"), use_underline=True)
        error_dismiss_button.connect("clicked", self._close_button_callback)

        self.error_box.add(error_label)
        self.error_box.add(error_dismiss_button)
        self.error_box.show_all()
        self.box.add(self.error_box)

    def set_text(self, text):
        self.log.get_buffer().set_text(text)

    def __init__(
            self, transient_for,
            headline=_("Exporting..."),
            width=150,
            height=80
    ):
        Gtk.Dialog.__init__(self, headline, transient_for, 0)
        self.set_default_size(width, height)

        self.label = CenterLabel()

        self.spinner = Gtk.Spinner()

        # Scrollable log window:
        self.log = Gtk.TextView()
        self.log.set_editable(False)
        # self.log.set_cursor_visible(False)
        if Gtk.get_minor_version() >= 16:
            self.log.set_monospace(True)
        else:
            self.log.override_font(
                Pango.font_description_from_string("monospace")
            )
        self.log.set_wrap_mode(Gtk.WrapMode.CHAR)
        #
        self.scrolled_window = Gtk.ScrolledWindow(expand=True)
        self.scrolled_window.set_margin_bottom(5)
        self.scrolled_window.add(self.log)
        #
        adj = self.scrolled_window.get_vadjustment()
        adj.connect(
            'changed',
            lambda adj: adj.set_value(adj.get_upper() - adj.get_page_size())
        )

        self.options_box = Gtk.Box(
            orientation=Gtk.Orientation.VERTICAL, spacing=5
        )
        self.options_box.set_margin_bottom(10)

        self.apply_button = Gtk.Button(label=_("_Apply"), use_underline=True)
        self.apply_button.connect("clicked", lambda x: self.do_export())

        self.error_box = Gtk.Box(
            orientation=Gtk.Orientation.VERTICAL, spacing=5
        )
        self.error_box.set_margin_bottom(10)

        self.box = self.get_content_area()
        self.box.set_margin_left(5)
        self.box.set_margin_right(5)
        self.box.set_spacing(5)
        self.box.add(self.label)

        self.show_all()

    def do_export(self):
        self.box.remove(self.options_box)
        self.box.remove(self.apply_button)
        self.box.add(self.spinner)
        self.box.add(self.scrolled_window)
        self.scrolled_window.set_size_request(-1, 200)
        self.scrolled_window.show_all()
        self.spinner.show()
        self.spinner.start()

        def update_ui(text):
            self.set_text(text)

        def ui_done():
            self.destroy()

        def ui_error():
            self.show_error()

        def do_export():
            self.label.set_text(_("Please wait while\nnew colorscheme will be created"))
            captured_log = ""
            proc = subprocess.Popen(
                self.command,
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT
            )
            for line in iter(proc.stdout.readline, b''):
                captured_log += line.decode("utf-8")
                GLib.idle_add(update_ui, captured_log)
            proc.communicate(timeout=self.timeout)
            if proc.returncode == 0:
                GLib.idle_add(ui_done)
            else:
                GLib.idle_add(ui_error)

        thread = Thread(target=do_export)
        thread.daemon = True
        thread.start()


class FileBasedExportDialog(ExportDialog):

    theme_name = None
    temp_theme_path = None

    def __init__(self, transient_for, colorscheme, theme_name, **kwargs):
        super().__init__(transient_for=transient_for, **kwargs)
        self.theme_name = 'oomox-' + theme_name.split('/')[-1]
        colorscheme_copy = generate_terminal_colors_for_oomox(colorscheme)
        self.temp_theme_path = save_colorscheme(
            preset_name=theme_name,
            colorscheme=colorscheme_copy,
            path=tempfile.mkstemp()[1]
        )

    def __del__(self):
        os.remove(self.temp_theme_path)


def export_terminal_theme(transient_for, colorscheme):
    dialog = ExportDialog(
        transient_for=transient_for,
        headline=_("Terminal colorscheme"),
        height=440
    )
    dialog.box.add(dialog.scrolled_window)
    dialog.scrolled_window.show_all()
    dialog.label.set_text(_('Paste this colorscheme to your ~/.Xresources'))
    try:
        term_colorscheme = generate_xrdb_theme_from_oomox(colorscheme)
        xresources_theme = generate_xresources(term_colorscheme)
    except Exception as exc:
        dialog.set_text(exc)
        dialog.show_error()
    else:
        dialog.set_text(xresources_theme)


OPTION_GTK2_HIDPI = 'gtk2_hidpi'


class GtkThemeExportDialog(FileBasedExportDialog):

    config_name = 'gtk_theme'

    def _create_option_checkbox_callback(self, option_id):
        def callback(widget):
            self.export_config[option_id] = widget.get_active()
        return callback

    def __init__(  # pylint: disable=too-many-arguments
            self, transient_for, colorscheme, theme_name,
            add_options=None, override_options=None,
            **kwargs
    ):
        super().__init__(
            transient_for=transient_for, colorscheme=colorscheme, theme_name=theme_name,
            **kwargs
        )

        export_options = override_options or {
            OPTION_GTK2_HIDPI: {
                'default': False,
                'display_name': _("Generate 2x scaled (_HiDPI) assets for GTK+2"),
            },
        }
        if add_options:
            export_options.update(add_options)
        self.export_config = ExportConfig(
            config_name=self.config_name,
            default_config={
                option_name: option['default']
                for option_name, option in export_options.items()
            }
        )

        self.label.set_text(_("Please choose theme export options:"))

        for option_name, option in export_options.items():
            checkbox = \
                Gtk.CheckButton.new_with_mnemonic(
                    option['display_name']
                )
            checkbox.connect(
                "toggled", self._create_option_checkbox_callback(option_name)
            )
            checkbox.set_active(
                self.export_config[option_name]
            )
            self.options_box.add(checkbox)

        self.box.add(self.options_box)
        self.options_box.show_all()
        self.box.add(self.apply_button)
        self.apply_button.show()

    def do_export(self):
        self.export_config.save()
        super().do_export()



import os

from oomox_gui.plugin_api import OomoxImportPlugin
from oomox_gui.i18n import _


PLUGIN_DIR = os.path.dirname(os.path.realpath(__file__))


class Plugin(OomoxImportPlugin):

    name = 'import_base16'
    display_name = 'Base16 YML format import'
    file_extensions = ('.yml', )

    theme_model_import = [
        {
            'display_name': _('Import Base16 colortheme:'),
            'type': 'separator',
            'value_filter': {
                'FROM_PLUGIN': 'import_base16',
            },
        },
        {
            'key': 'BASE16_GENERATE_DARK',
            'type': 'bool',
            'fallback_value': False,
            'display_name': _('Generate dark GUI variant'),
            'reload_theme': True,
        },
        {
            'display_name': _('Theme options:'),
            'type': 'separator',
            'value_filter': {
                'FROM_PLUGIN': 'import_base16',
            },
        },
    ]

    default_theme = {
        "TERMINAL_THEME_MODE": "manual",
    }
    translation_common = {
        "NAME": "scheme",

        "TERMINAL_COLOR0": "base00",
        "TERMINAL_COLOR1": "base08",
        "TERMINAL_COLOR2": "base0b",
        "TERMINAL_COLOR3": "base09",
        "TERMINAL_COLOR4": "base0d",
        "TERMINAL_COLOR5": "base0e",
        "TERMINAL_COLOR6": "base0c",
        "TERMINAL_COLOR7": "base05",
        "TERMINAL_COLOR8": "base02",
        "TERMINAL_COLOR9": "base08",  # @TODO: lighter
        "TERMINAL_COLOR10": "base0b",  # @TODO: lighter
        "TERMINAL_COLOR11": "base0a",
        "TERMINAL_COLOR12": "base0d",  # @TODO: lighter
        "TERMINAL_COLOR13": "base0e",  # @TODO: lighter
        "TERMINAL_COLOR14": "base0c",  # @TODO: lighter
        "TERMINAL_COLOR15": "base07",

        "ICONS_LIGHT_FOLDER": "base0c",
        "ICONS_LIGHT": "base0c",
        "ICONS_MEDIUM": "base0d",
        "ICONS_DARK": "base03",
    }
    translation_light = {
        "BG": "base05",
        "FG": "base00",
        "MENU_BG": "base04",
        "MENU_FG": "base01",
        "SEL_BG": "base0d",
        "SEL_FG": "base00",
        "TXT_BG": "base06",
        "TXT_FG": "base01",
        "BTN_BG": "base03",
        "BTN_FG": "base07",
        "HDR_BTN_BG": "base05",
        "HDR_BTN_FG": "base01",

        "TERMINAL_BACKGROUND": "base01",
        "TERMINAL_FOREGROUND": "base06",
        "TERMINAL_ACCENT_COLOR": "base0d",
    }
    translation_dark = {
        "BG": "base01",
        "FG": "base06",
        "MENU_BG": "base00",
        "MENU_FG": "base05",
        "SEL_BG": "base08",
        "SEL_FG": "base00",
        "TXT_BG": "base02",
        "TXT_FG": "base07",
        "BTN_BG": "base00",
        "BTN_FG": "base05",
        "HDR_BTN_BG": "base01",
        "HDR_BTN_FG": "base05",

        "TERMINAL_COLOR8": "base01",
        "TERMINAL_BACKGROUND": "base02",
        "TERMINAL_FOREGROUND": "base07",
        "TERMINAL_ACCENT_COLOR": "base08",
    }

    def read_colorscheme_from_path(self, preset_path):
        from oomox_gui.theme_model import THEME_MODEL_BY_KEY

        base16_theme = {}
        with open(preset_path) as preset_file:
            for line in preset_file.readlines():
                try:
                    key, value, *_rest = line.split()
                    key = key.rstrip(':').lower()
                    value = value.strip('\'"').lower()
                    base16_theme[key] = value
                except Exception:
                    pass

        oomox_theme = {}
        oomox_theme.update(self.default_theme)
        translation = {}
        translation.update(self.translation_common)
        if THEME_MODEL_BY_KEY.get('BASE16_GENERATE_DARK', {}).get('fallback_value'):
            translation.update(self.translation_dark)
        else:
            translation.update(self.translation_light)
        for oomox_key, base16_key in translation.items():
            if base16_key in base16_theme:
                oomox_theme[oomox_key] = base16_theme[base16_key]
        return oomox_theme



import os
import sys
import importlib


def mkdir_p(path):
    if os.path.isdir(path):
        return
    os.makedirs(path)


def ls_r(path):
    return [
        os.path.join(files[0], file)
        for files in os.walk(path, followlinks=True) for file in files[2]
    ]


def get_plugin_module(name, path):
    if sys.version_info.minor >= 5:
        spec = importlib.util.spec_from_file_location(name, path)  # pylint: disable=no-member
        module = importlib.util.module_from_spec(spec)  # pylint: disable=no-member
        spec.loader.exec_module(module)
    else:
        loader = importlib.machinery.SourceFileLoader(name, path)
        module = loader.load_module()  # pylint: disable=deprecated-method,no-value-for-parameter
    return module



import os

from oomox_gui.export_common import GtkThemeExportDialog, OPTION_GTK2_HIDPI
from oomox_gui.plugin_api import OomoxThemePlugin
from oomox_gui.i18n import _


PLUGIN_DIR = os.path.dirname(os.path.realpath(__file__))
THEME_DIR = os.path.join(PLUGIN_DIR, "materia-theme/")


class MateriaThemeExportDialog(GtkThemeExportDialog):
    timeout = 1000

    def do_export(self):
        self.command = [
            "bash",
            os.path.join(THEME_DIR, "change_color.sh"),
            "--hidpi", str(self.export_config[OPTION_GTK2_HIDPI]),
            "--output", self.theme_name,
            self.temp_theme_path,
        ]
        super().do_export()


class Plugin(OomoxThemePlugin):

    name = 'materia'
    display_name = 'Materia'
    description = '(GTK+2, GTK+3, Cinnamon, Gnome Shell, Metacity, Unity, Xfwm)'
    export_dialog = MateriaThemeExportDialog
    gtk_preview_dir = os.path.join(PLUGIN_DIR, "gtk_preview_css/")
    preview_sizes = {
        OomoxThemePlugin.PreviewImageboxesNames.CHECKBOX.name: 24,
    }

    enabled_keys_gtk = [
        'BG',
        'FG',
        'MENU_BG',
        'MENU_FG',
        'SEL_BG',
        'SEL_FG',
        'ACCENT_BG',
        'TXT_BG',
        'BTN_BG',
        'BTN_FG',
    ]
    enabled_keys_options = [
        'ROUNDNESS',
    ]

    theme_model_gtk = []

    theme_model_options = [
        {
            'key': 'GTK3_GENERATE_DARK',
            'type': 'bool',
            'fallback_value': True,
            'display_name': _('(GTK3) Add dark variant'),
        },
        {
            'key': 'MATERIA_STYLE_COMPACT',
            'type': 'bool',
            'fallback_value': True,
            'display_name': _('Compact style'),
        },
    ]

    theme_model_other = [
        {
            'key': 'GNOME_SHELL_PANEL_OPACITY',
            'type': 'float',
            'fallback_value': 0.6,
            'max_value': 1.0,
            'display_name': _('(Gnome Shell) Panel opacity'),
        },
    ]

    def preview_before_load_callback(self, preview_object, colorscheme):
        colorscheme["TXT_FG"] = colorscheme["FG"]
        colorscheme["WM_BORDER_FOCUS"] = colorscheme["MENU_BG"]
        colorscheme["WM_BORDER_UNFOCUS"] = colorscheme["BTN_BG"]
        colorscheme["HDR_BTN_FG"] = colorscheme["MENU_FG"]
        colorscheme["HDR_BTN_BG"] = colorscheme["MENU_BG"]
        colorscheme["GRADIENT"] = 0
        preview_object.WM_BORDER_WIDTH = 0



from PIL import Image


SMOKE, WEED, EVERYDAY = 0, 1, 2


RAICHU = 200
RAIKOU = 1000
DEOXYS = 10000


def swablu(salamence, golduck, lileep=WEED + EVERYDAY):
    # pylint: disable=too-many-branches
    azumarill = salamence
    beautifly = len(azumarill)

    espeon = beautifly
    umbreon = SMOKE
    while espeon > umbreon:

        espeon = len(azumarill)
        kangaskhan = EVERYDAY
        while espeon > umbreon and kangaskhan < len(azumarill) and beautifly > WEED:

            espeon = len(azumarill)
            kangaskhan = EVERYDAY

            while beautifly > WEED:
                jynx = azumarill[beautifly - WEED]
                hitmonlee = azumarill[beautifly - kangaskhan]
                ledian = jynx[WEED]
                skitty = hitmonlee[WEED]
                if ((abs(ledian[SMOKE] - skitty[SMOKE]) <= lileep
                     and abs(ledian[WEED] - skitty[WEED]) <= lileep
                     and abs(ledian[EVERYDAY] - skitty[EVERYDAY]) <= lileep)):
                    ditto = jynx[SMOKE]
                    charizard = hitmonlee[SMOKE]
                    treecko = ditto + charizard
                    if ditto > charizard:
                        azumarill[beautifly - WEED] = [treecko, ledian]
                    else:
                        azumarill[beautifly - kangaskhan] = [
                            treecko, skitty
                        ]
                    if ditto > charizard:
                        del azumarill[beautifly - kangaskhan]
                    else:
                        del azumarill[beautifly - WEED]
                beautifly -= WEED
                kangaskhan += WEED
                # if beautifly < WEED:
                #     break
                if kangaskhan > len(azumarill):
                    break

            umbreon = len(azumarill)
            if umbreon > golduck:
                salamence = azumarill
            else:
                break

        umbreon = len(azumarill)
        if umbreon > golduck:
            salamence = azumarill
        else:
            break

    return salamence


def mewtwo(caterpie,
           golduck=16,
           lileep=WEED + WEED + EVERYDAY,
           machoke=WEED,
           pelipper=SMOKE):

    if DEOXYS:
        salamence = list(reversed(caterpie))[:DEOXYS]
    else:
        salamence = list(reversed(caterpie))

    metagross = SMOKE
    nidorina = SMOKE
    persian = lileep
    dunsparce = None
    if pelipper:
        for pikachu in [SMOKE, WEED, EVERYDAY]:
            def flareon(venonat, slowpoke=pikachu):
                return venonat[WEED][slowpoke]
            salamence.sort(key=flareon)
            for _meowth in range(SMOKE, machoke):
                salamence = swablu(
                    salamence,
                    golduck,
                    lileep=lileep
                )
    while (len(salamence) > golduck) and (persian < RAICHU):

        salamence.sort(
            key=lambda venonat: venonat[SMOKE],
            reverse=WEED
        )
        salamence = swablu(salamence, golduck, persian)

        if dunsparce == len(salamence):
            nidorina = nidorina + WEED
        if nidorina > WEED + WEED + EVERYDAY:  # after what tries increase lileep
            nidorina = SMOKE
            metagross += WEED
            persian += WEED
        if metagross > RAIKOU:  # after what lileep-increases raise an error
            break
        dunsparce = len(salamence)

    return wobbuffet(salamence, xatu=SMOKE)


def wobbuffet(natu, xatu=WEED):
    return sorted(natu, key=lambda venonat: venonat[SMOKE], reverse=xatu)


def delibird(girafarig):
    return "{:02x}".format(max(0, min(255, int(girafarig))))


def jumpluff(caterpie):
    return ''.join([delibird(meowth) for meowth in caterpie])


def jolteon(smeargle, bulbasaur):
    venonat = min(bulbasaur, smeargle.size[SMOKE])
    hitmontop = int(
        round(smeargle.size[WEED] / (smeargle.size[SMOKE] / venonat)))
    print((venonat, hitmontop))
    smeargle = smeargle.convert('RGB')
    skarmory = smeargle.resize((venonat, hitmontop), )
    parasect = skarmory.getcolors(
        maxcolors=skarmory.size[SMOKE] * skarmory.size[WEED])
    return parasect


def get_hex_palette(image_path, use_whole_palette=False, accuracy=48, quality=400):
    smeargle = Image.open(image_path)
    whirlipede = jolteon(smeargle, quality)
    if not use_whole_palette:
        whirlipede = wobbuffet(mewtwo(
            whirlipede,
            golduck=accuracy,
            lileep=WEED + WEED + EVERYDAY,
            machoke=EVERYDAY,
            pelipper=SMOKE
        ))
    hex_palette = []
    for _, caterpie in whirlipede:
        if len(caterpie) > EVERYDAY+WEED:
            caterpie = caterpie[:EVERYDAY+WEED]
        hex_palette.append(jumpluff(caterpie))
    return hex_palette



import os


SCRIPT_DIR = os.path.dirname(os.path.realpath(__file__))

OOMOX_ROOT_DIR = os.path.join(SCRIPT_DIR, "../")

COLORS_DIR = os.path.join(
    OOMOX_ROOT_DIR, "colors/"
)
PLUGINS_DIR = os.path.join(
    OOMOX_ROOT_DIR, "plugins/"
)
TERMINAL_TEMPLATE_DIR = os.path.join(
    OOMOX_ROOT_DIR, "terminal_templates/"
)


USER_CONFIG_DIR = os.path.join(
    os.environ.get(
        "XDG_CONFIG_HOME",
        os.path.join(os.environ.get("HOME"), ".config/")
    ),
    "oomox/"
)
USER_COLORS_DIR = os.path.join(
    USER_CONFIG_DIR, "colors/"
)
USER_PLUGINS_DIR = os.path.join(
    USER_CONFIG_DIR, "plugins/"
)
USER_PALETTE_PATH = os.path.join(
    USER_CONFIG_DIR, "recent_palette.json"
)
USER_EXPORT_CONFIG_DIR = os.path.join(
    USER_CONFIG_DIR, "export_config/"
)


FALLBACK_COLOR = "F33333"



from .theme_model import THEME_MODEL
from .color import get_random_theme_color
from .xrdb import XrdbCache
from .plugin_loader import IMPORT_PLUGINS


def str_to_bool(value):
    return value.lower() == 'true'


def parse_theme_color_value(result_value):
    if not result_value:
        return None
    if result_value == 'random_color':
        result_value = get_random_theme_color()
    elif result_value.startswith('xrdb.'):
        xrdb_color = XrdbCache.get().get(result_value.replace('xrdb.', ''))
        if xrdb_color and xrdb_color.startswith('#'):
            result_value = xrdb_color.replace('#', '')
    return result_value


def parse_theme_value(theme_value, colorscheme):  # pylint: disable=too-many-branches
    result_value = colorscheme.get(theme_value['key'])
    fallback_key = theme_value.get('fallback_key')
    fallback_value = theme_value.get('fallback_value')
    fallback_function = theme_value.get('fallback_function')

    if result_value is None:
        if fallback_value is not None:
            result_value = fallback_value
        elif fallback_key:
            result_value = colorscheme[fallback_key]
        elif fallback_function:
            result_value = fallback_function(colorscheme)

    value_type = theme_value['type']
    if value_type == 'color':
        result_value = parse_theme_color_value(result_value)
    if value_type == 'bool':
        if isinstance(result_value, str):
            result_value = str_to_bool(result_value)
    elif value_type == 'int':
        result_value = int(result_value)
    elif value_type == 'float':
        result_value = float(result_value)
    elif value_type == 'options':
        available_options = [option['value'] for option in theme_value['options']]
        if result_value not in available_options:
            if fallback_value in available_options:
                result_value = fallback_value
            else:
                result_value = available_options[0]

    return result_value


def read_colorscheme_from_path(preset_path):
    colorscheme = {}

    from_plugin = None
    for plugin_name, plugin in IMPORT_PLUGINS.items():
        for file_extension in plugin.file_extensions:
            if preset_path.lower().endswith(file_extension):
                colorscheme = plugin.read_colorscheme_from_path(preset_path)
                from_plugin = plugin_name
                break
    if not colorscheme:
        theme_keys = [item['key'] for item in THEME_MODEL if 'key' in item]

        theme_keys.append('NOGUI')

        with open(preset_path) as file_object:
            for line in file_object.readlines():
                parsed_line = line.strip().split('=')
                key = parsed_line[0]
                if not key.startswith("#"):
                    if key in theme_keys and len(parsed_line) > 1:
                        colorscheme[key] = parsed_line[1]

    for theme_model_item in THEME_MODEL:
        key = theme_model_item.get('key')
        if not key:
            continue
        colorscheme[key] = parse_theme_value(theme_model_item, colorscheme)
    if from_plugin:
        colorscheme['FROM_PLUGIN'] = from_plugin

    XrdbCache.clear()
    return colorscheme



import os
from collections import defaultdict
from itertools import groupby

from .config import COLORS_DIR, USER_COLORS_DIR
from .helpers import ls_r, mkdir_p


def get_presets():
    file_paths = [
        {
            "name": "".join(
                path.startswith(COLORS_DIR) and path.rsplit(COLORS_DIR) or
                path.rsplit(USER_COLORS_DIR)
            ),
            "path": path,
            "default": is_default,
        }
        for paths, is_default in (
            (ls_r(USER_COLORS_DIR), False),
            (ls_r(COLORS_DIR), True)
        )
        for path in paths
    ]
    result = defaultdict(list)
    for _key, group in groupby(file_paths, lambda x: x['name'].split('/')[0]):
        group = sorted(list(group), key=lambda x: x['name'])
        display_name = group[0]['name']
        if display_name in result:
            display_name = display_name + " (default)"
        result[display_name] = group
    return dict(result)


def get_user_theme_path(user_theme_name):
    return os.path.join(USER_COLORS_DIR, user_theme_name)


def save_colorscheme(preset_name, colorscheme, path=None):
    colorscheme["NAME"] = preset_name
    path = path or get_user_theme_path(preset_name)
    if not os.path.exists(path):
        mkdir_p(os.path.dirname(path))
    with open(path, 'w') as file_object:
        for key, value in sorted(colorscheme.items()):
            if (
                    key not in ('NOGUI', )
            ) and (
                not key.startswith('_')
            ) and (
                value is not None
            ):
                file_object.write("{}={}\n".format(
                    key, value
                ))
    return path


def remove_colorscheme(preset_name):
    path = os.path.join(USER_COLORS_DIR, preset_name)
    os.remove(path)


def is_user_colorscheme(preset_path):
    return preset_path.startswith(USER_COLORS_DIR)


def is_colorscheme_exists(preset_path):
    return os.path.exists(preset_path)
